# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""MAD4PG trainer implementation."""
from typing import Any, Dict, List, Union

import sonnet as snt
import tensorflow as tf
import tree
from acme.tf import losses
from acme.tf import utils as tf2_utils
from acme.utils import counting, loggers

from mava.components.tf.losses.sequence import recurrent_n_step_critic_loss
from mava.systems.tf.maddpg.training import (
    MADDPGBaseRecurrentTrainer,
    MADDPGBaseTrainer,
    MADDPGCentralisedRecurrentTrainer,
    MADDPGCentralisedTrainer,
    MADDPGDecentralisedRecurrentTrainer,
    MADDPGDecentralisedTrainer,
    MADDPGStateBasedRecurrentTrainer,
    MADDPGStateBasedTrainer,
)
from mava.utils import training_utils as train_utils

train_utils.set_growing_gpu_memory()


class MAD4PGBaseTrainer(MADDPGBaseTrainer):
    """MAD4PG trainer.
    This is the trainer component of a MAD4PG system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        target_policy_networks: Dict[str, snt.Module],
        target_critic_networks: Dict[str, snt.Module],
        policy_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        critic_optimizer: snt.Optimizer,
        discount: float,
        target_averaging: bool,
        target_update_period: int,
        target_update_rate: float,
        dataset: tf.data.Dataset,
        observation_networks: Dict[str, snt.Module],
        target_observation_networks: Dict[str, snt.Module],
        shared_weights: bool = False,
        max_gradient_norm: float = None,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
    ):
        """Initializes the learner.
        Args:
          policy_network: the online (optimized) policy.
          critic_network: the online critic.
          target_policy_network: the target policy (which lags behind the online
            policy).
          target_critic_network: the target critic.
          discount: discount to use for TD updates.
          target_update_period: number of learner steps to perform before updating
            the target networks.
          dataset: dataset to learn from, whether fixed or from a replay buffer
            (see `acme.datasets.reverb.make_dataset` documentation).
          observation_network: an optional online network to process observations
            before the policy and the critic.
          target_observation_network: the target observation network.
          policy_optimizer: the optimizer to be applied to the DPG (policy) loss.
          critic_optimizer: the optimizer to be applied to the critic loss.
          clipping: whether to clip gradients by global norm.
          counter: counter object used to keep track of steps.
          logger: logger object to be used by learner.
          checkpoint: boolean indicating whether to checkpoint the learner.
        """

        super().__init__(
            agents=agents,
            agent_types=agent_types,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            target_policy_networks=target_policy_networks,
            target_critic_networks=target_critic_networks,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            discount=discount,
            target_averaging=target_averaging,
            target_update_period=target_update_period,
            target_update_rate=target_update_rate,
            dataset=dataset,
            observation_networks=observation_networks,
            target_observation_networks=target_observation_networks,
            shared_weights=shared_weights,
            max_gradient_norm=max_gradient_norm,
            counter=counter,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
        )

    # Forward pass that calculates loss.
    def _forward(self, inputs: Any) -> None:
        # Unpack input data as follows:
        # o_tm1 = dictionary of observations one for each agent
        # a_tm1 = dictionary of actions taken from obs in o_tm1
        # e_tm1 [Optional] = extra data for timestep t-1
        # that the agents persist in replay.
        # r_t = dictionary of rewards or rewards sequences
        #   (if using N step transitions) ensuing from actions a_tm1
        # d_t = environment discount ensuing from actions a_tm1.
        #   This discount is applied to future rewards after r_t.
        # o_t = dictionary of next observations or next observation sequences
        # e_t [Optional] = extra data for timestep t that the agents persist in replay.
        o_tm1, a_tm1, e_tm1, r_t, d_t, o_t, e_t = inputs.data

        # Do forward passes through the networks and calculate the losses
        self.policy_losses = {}
        self.critic_losses = {}
        with tf.GradientTape(persistent=True) as tape:
            o_tm1_trans, o_t_trans = self._transform_observations(o_tm1, o_t)
            a_t = self._target_policy_actions(o_t_trans)

            for agent in self._agents:
                agent_key = self.agent_net_keys[agent]

                # Get critic feed
                o_tm1_feed, o_t_feed, a_tm1_feed, a_t_feed = self._get_critic_feed(
                    o_tm1_trans=o_tm1_trans,
                    o_t_trans=o_t_trans,
                    a_tm1=a_tm1,
                    a_t=a_t,
                    e_tm1=e_tm1,
                    e_t=e_t,
                    agent=agent,
                )

                # Critic learning.
                q_tm1 = self._critic_networks[agent_key](o_tm1_feed, a_tm1_feed)
                q_t = self._target_critic_networks[agent_key](o_t_feed, a_t_feed)

                # Cast the additional discount to match the environment discount dtype.
                discount = tf.cast(self._discount, dtype=d_t[agent].dtype)

                # Critic loss.
                critic_loss = losses.categorical(
                    q_tm1, r_t[agent], discount * d_t[agent], q_t
                )
                self.critic_losses[agent] = tf.reduce_mean(critic_loss, axis=0)
                # Actor learning.
                o_t_agent_feed = o_t_trans[agent]
                dpg_a_t = self._policy_networks[agent_key](o_t_agent_feed)

                # Get dpg actions
                dpg_a_t_feed = self._get_dpg_feed(a_t, dpg_a_t, agent)

                # Get dpg Q values.
                dpg_z_t = self._critic_networks[agent_key](o_t_feed, dpg_a_t_feed)
                dpg_q_t = dpg_z_t.mean()

                # Actor loss. If clipping is true use dqda clipping and clip the norm.
                dqda_clipping = 1.0 if self._max_gradient_norm is not None else None
                clip_norm = True if self._max_gradient_norm is not None else False

                policy_loss = losses.dpg(
                    dpg_q_t,
                    dpg_a_t,
                    tape=tape,
                    dqda_clipping=dqda_clipping,
                    clip_norm=clip_norm,
                )
                self.policy_losses[agent] = tf.reduce_mean(policy_loss, axis=0)
        self.tape = tape


class MAD4PGDecentralisedTrainer(MAD4PGBaseTrainer, MADDPGDecentralisedTrainer):
    """MAD4PG trainer.
    This is the trainer component of a MAD4PG system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        target_policy_networks: Dict[str, snt.Module],
        target_critic_networks: Dict[str, snt.Module],
        policy_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        critic_optimizer: snt.Optimizer,
        discount: float,
        target_averaging: bool,
        target_update_period: int,
        target_update_rate: float,
        dataset: tf.data.Dataset,
        observation_networks: Dict[str, snt.Module],
        target_observation_networks: Dict[str, snt.Module],
        shared_weights: bool = False,
        max_gradient_norm: float = None,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
    ):
        """Initializes the learner.
        Args:
          policy_network: the online (optimized) policy.
          critic_network: the online critic.
          target_policy_network: the target policy (which lags behind the online
            policy).
          target_critic_network: the target critic.
          discount: discount to use for TD updates.
          target_update_period: number of learner steps to perform before updating
            the target networks.
          dataset: dataset to learn from, whether fixed or from a replay buffer
            (see `acme.datasets.reverb.make_dataset` documentation).
          observation_network: an optional online network to process observations
            before the policy and the critic.
          target_observation_network: the target observation network.
          policy_optimizer: the optimizer to be applied to the DPG (policy) loss.
          critic_optimizer: the optimizer to be applied to the critic loss.
          clipping: whether to clip gradients by global norm.
          counter: counter object used to keep track of steps.
          logger: logger object to be used by learner.
          checkpoint: boolean indicating whether to checkpoint the learner.
        """

        super().__init__(
            agents=agents,
            agent_types=agent_types,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            target_policy_networks=target_policy_networks,
            target_critic_networks=target_critic_networks,
            discount=discount,
            target_averaging=target_averaging,
            target_update_period=target_update_period,
            target_update_rate=target_update_rate,
            dataset=dataset,
            observation_networks=observation_networks,
            target_observation_networks=target_observation_networks,
            shared_weights=shared_weights,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            max_gradient_norm=max_gradient_norm,
            counter=counter,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
        )


class MAD4PGCentralisedTrainer(MAD4PGBaseTrainer, MADDPGCentralisedTrainer):
    """MAD4PG trainer.
    This is the trainer component of a MAD4PG system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        target_policy_networks: Dict[str, snt.Module],
        target_critic_networks: Dict[str, snt.Module],
        policy_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        critic_optimizer: snt.Optimizer,
        discount: float,
        target_averaging: bool,
        target_update_period: int,
        target_update_rate: float,
        dataset: tf.data.Dataset,
        observation_networks: Dict[str, snt.Module],
        target_observation_networks: Dict[str, snt.Module],
        shared_weights: bool = False,
        max_gradient_norm: float = None,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
    ):
        """Initializes the learner.
        Args:
          policy_network: the online (optimized) policy.
          critic_network: the online critic.
          target_policy_network: the target policy (which lags behind the online
            policy).
          target_critic_network: the target critic.
          discount: discount to use for TD updates.
          target_update_period: number of learner steps to perform before updating
            the target networks.
          dataset: dataset to learn from, whether fixed or from a replay buffer
            (see `acme.datasets.reverb.make_dataset` documentation).
          observation_network: an optional online network to process observations
            before the policy and the critic.
          target_observation_network: the target observation network.
          policy_optimizer: the optimizer to be applied to the DPG (policy) loss.
          critic_optimizer: the optimizer to be applied to the critic loss.
          clipping: whether to clip gradients by global norm.
          counter: counter object used to keep track of steps.
          logger: logger object to be used by learner.
          checkpoint: boolean indicating whether to checkpoint the learner.
        """

        super().__init__(
            agents=agents,
            agent_types=agent_types,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            target_policy_networks=target_policy_networks,
            target_critic_networks=target_critic_networks,
            discount=discount,
            target_averaging=target_averaging,
            target_update_period=target_update_period,
            target_update_rate=target_update_rate,
            dataset=dataset,
            observation_networks=observation_networks,
            target_observation_networks=target_observation_networks,
            shared_weights=shared_weights,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            max_gradient_norm=max_gradient_norm,
            counter=counter,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
        )


class MAD4PGStateBasedTrainer(MAD4PGBaseTrainer, MADDPGStateBasedTrainer):
    """MAD4PG trainer.
    This is the trainer component of a MAD4PG system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        target_policy_networks: Dict[str, snt.Module],
        target_critic_networks: Dict[str, snt.Module],
        policy_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        critic_optimizer: snt.Optimizer,
        discount: float,
        target_averaging: bool,
        target_update_period: int,
        target_update_rate: float,
        dataset: tf.data.Dataset,
        observation_networks: Dict[str, snt.Module],
        target_observation_networks: Dict[str, snt.Module],
        shared_weights: bool = False,
        max_gradient_norm: float = None,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
    ):
        """Initializes the learner.
        Args:
          policy_network: the online (optimized) policy.
          critic_network: the online critic.
          target_policy_network: the target policy (which lags behind the online
            policy).
          target_critic_network: the target critic.
          discount: discount to use for TD updates.
          target_update_period: number of learner steps to perform before updating
            the target networks.
          dataset: dataset to learn from, whether fixed or from a replay buffer
            (see `acme.datasets.reverb.make_dataset` documentation).
          observation_network: an optional online network to process observations
            before the policy and the critic.
          target_observation_network: the target observation network.
          policy_optimizer: the optimizer to be applied to the DPG (policy) loss.
          critic_optimizer: the optimizer to be applied to the critic loss.
          clipping: whether to clip gradients by global norm.
          counter: counter object used to keep track of steps.
          logger: logger object to be used by learner.
          checkpoint: boolean indicating whether to checkpoint the learner.
        """

        super().__init__(
            agents=agents,
            agent_types=agent_types,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            target_policy_networks=target_policy_networks,
            target_critic_networks=target_critic_networks,
            discount=discount,
            target_averaging=target_averaging,
            target_update_period=target_update_period,
            target_update_rate=target_update_rate,
            dataset=dataset,
            observation_networks=observation_networks,
            target_observation_networks=target_observation_networks,
            shared_weights=shared_weights,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            max_gradient_norm=max_gradient_norm,
            counter=counter,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
        )


class MAD4PGBaseRecurrentTrainer(MADDPGBaseRecurrentTrainer):
    """MAD4PG trainer.
    This is the trainer component of a MAD4PG system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        target_policy_networks: Dict[str, snt.Module],
        target_critic_networks: Dict[str, snt.Module],
        policy_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        critic_optimizer: snt.Optimizer,
        discount: float,
        target_averaging: bool,
        target_update_period: int,
        target_update_rate: float,
        dataset: tf.data.Dataset,
        observation_networks: Dict[str, snt.Module],
        target_observation_networks: Dict[str, snt.Module],
        shared_weights: bool = False,
        max_gradient_norm: float = None,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
        bootstrap_n: int = 10,
    ):
        """Initializes the learner.
        Args:
          policy_network: the online (optimized) policy.
          critic_network: the online critic.
          target_policy_network: the target policy (which lags behind the online
            policy).
          target_critic_network: the target critic.
          discount: discount to use for TD updates.
          target_update_period: number of learner steps to perform before updating
            the target networks.
          dataset: dataset to learn from, whether fixed or from a replay buffer
            (see `acme.datasets.reverb.make_dataset` documentation).
          observation_network: an optional online network to process observations
            before the policy and the critic.
          target_observation_network: the target observation network.
          policy_optimizer: the optimizer to be applied to the DPG (policy) loss.
          critic_optimizer: the optimizer to be applied to the critic loss.
          clipping: whether to clip gradients by global norm.
          counter: counter object used to keep track of steps.
          logger: logger object to be used by learner.
          checkpoint: boolean indicating whether to checkpoint the learner.
        """

        super().__init__(
            agents=agents,
            agent_types=agent_types,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            target_policy_networks=target_policy_networks,
            target_critic_networks=target_critic_networks,
            discount=discount,
            target_averaging=target_averaging,
            target_update_period=target_update_period,
            target_update_rate=target_update_rate,
            dataset=dataset,
            observation_networks=observation_networks,
            target_observation_networks=target_observation_networks,
            shared_weights=shared_weights,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            max_gradient_norm=max_gradient_norm,
            counter=counter,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
            bootstrap_n=bootstrap_n,
        )

    # Forward pass that calculates loss.
    def _forward(self, inputs: Any) -> None:
        # TODO: Update this forward function to work like MA-D4PG
        data = inputs.data

        # Note (dries): The unused variable is start_of_episodes.
        observations, actions, rewards, discounts, _, extras = (
            data.observations,
            data.actions,
            data.rewards,
            data.discounts,
            data.start_of_episode,
            data.extras,
        )

        # Get initial state for the LSTM from replay and
        # extract the first state in the sequence..
        core_state = tree.map_structure(lambda s: s[:, 0, :], extras["core_states"])
        target_core_state = tree.map_structure(tf.identity, core_state)

        # TODO (dries): Take out all the data_points that does not need
        #  to be processed here at the start. Therefore it does not have
        #  to be done later on and saves processing time.

        self.policy_losses: Dict[str, tf.Tensor] = {}
        self.critic_losses: Dict[str, tf.Tensor] = {}

        # Do forward passes through the networks and calculate the losses
        with tf.GradientTape(persistent=True) as tape:
            # Note (dries): We are assuming that only the policy network
            # is recurrent and not the observation network.
            obs_trans, target_obs_trans = self._transform_observations(observations)

            target_actions = self._target_policy_actions(
                target_obs_trans, target_core_state
            )

            for agent in self._agents:
                agent_key = self.agent_net_keys[agent]

                # Get critic feed
                (
                    obs_trans_feed,
                    target_obs_trans_feed,
                    action_feed,
                    target_actions_feed,
                ) = self._get_critic_feed(
                    obs_trans=obs_trans,
                    target_obs_trans=target_obs_trans,
                    actions=actions,
                    target_actions=target_actions,
                    extras=extras,
                    agent=agent,
                )

                # Critic learning.
                # Remove the last sequence step for the normal network
                obs_comb, dims = train_utils.combine_dim(obs_trans_feed)
                act_comb, _ = train_utils.combine_dim(action_feed)
                q_values = self._critic_networks[agent_key](obs_comb, act_comb)
                q_values.set_dimensions(dims)

                # Remove first sequence step for the target
                obs_comb, _ = train_utils.combine_dim(target_obs_trans_feed)
                act_comb, _ = train_utils.combine_dim(target_actions_feed)
                target_q_values = self._target_critic_networks[agent_key](
                    obs_comb, act_comb
                )
                target_q_values.set_dimensions(dims)

                # Cast the additional discount to match
                # the environment discount dtype.
                agent_discount = discounts[agent]
                discount = tf.cast(self._discount, dtype=agent_discount.dtype)

                # Critic loss.
                critic_loss = recurrent_n_step_critic_loss(
                    q_values,
                    target_q_values,
                    rewards[agent],
                    discount * agent_discount,
                    bootstrap_n=self._bootstrap_n,
                    loss_fn=losses.categorical,
                )
                self.critic_losses[agent] = tf.reduce_mean(critic_loss, axis=0)

                # Actor learning.
                obs_agent_feed = target_obs_trans[agent]
                # TODO (dries): Why is there an extra tuple?
                agent_core_state = core_state[agent][0]
                transposed_obs = tf2_utils.batch_to_sequence(obs_agent_feed)
                outputs, updated_states = snt.static_unroll(
                    self._policy_networks[agent_key],
                    transposed_obs,
                    agent_core_state,
                )

                dpg_actions = tf2_utils.batch_to_sequence(outputs)

                # Note (dries): This is done to so that losses.dpg can verify
                # using gradient.tape that there is a
                # gradient relationship between dpg_q_values and dpg_actions_comb.
                dpg_actions_comb, dim = train_utils.combine_dim(dpg_actions)

                # Note (dries): This seemingly useless line is important!
                # Don't remove it. See above note.
                dpg_actions = train_utils.extract_dim(dpg_actions_comb, dim)

                # Get dpg actions
                dpg_actions_feed = self._get_dpg_feed(
                    target_actions, dpg_actions, agent
                )

                # Get dpg Q values.
                obs_comb, _ = train_utils.combine_dim(target_obs_trans_feed)
                act_comb, _ = train_utils.combine_dim(dpg_actions_feed)
                dpg_z_values = self._critic_networks[agent_key](obs_comb, act_comb)
                dpg_q_values = dpg_z_values.mean()

                # Actor loss. If clipping is true use dqda clipping and clip the norm.
                dqda_clipping = 1.0 if self._max_gradient_norm is not None else None
                clip_norm = True if self._max_gradient_norm is not None else False

                policy_loss = losses.dpg(
                    dpg_q_values,
                    dpg_actions_comb,
                    tape=tape,
                    dqda_clipping=dqda_clipping,
                    clip_norm=clip_norm,
                )
                self.policy_losses[agent] = tf.reduce_mean(policy_loss, axis=0)
        self.tape = tape


class MAD4PGDecentralisedRecurrentTrainer(
    MAD4PGBaseRecurrentTrainer, MADDPGDecentralisedRecurrentTrainer
):
    """MAD4PG trainer.
    This is the trainer component of a MAD4PG system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        target_policy_networks: Dict[str, snt.Module],
        target_critic_networks: Dict[str, snt.Module],
        policy_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        critic_optimizer: snt.Optimizer,
        discount: float,
        target_averaging: bool,
        target_update_period: int,
        target_update_rate: float,
        dataset: tf.data.Dataset,
        observation_networks: Dict[str, snt.Module],
        target_observation_networks: Dict[str, snt.Module],
        shared_weights: bool = False,
        max_gradient_norm: float = None,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
        bootstrap_n: int = 10,
    ):
        """Initializes the learner.
        Args:
          policy_network: the online (optimized) policy.
          critic_network: the online critic.
          target_policy_network: the target policy (which lags behind the online
            policy).
          target_critic_network: the target critic.
          discount: discount to use for TD updates.
          target_update_period: number of learner steps to perform before updating
            the target networks.
          dataset: dataset to learn from, whether fixed or from a replay buffer
            (see `acme.datasets.reverb.make_dataset` documentation).
          observation_network: an optional online network to process observations
            before the policy and the critic.
          target_observation_network: the target observation network.
          policy_optimizer: the optimizer to be applied to the DPG (policy) loss.
          critic_optimizer: the optimizer to be applied to the critic loss.
          clipping: whether to clip gradients by global norm.
          counter: counter object used to keep track of steps.
          logger: logger object to be used by learner.
          checkpoint: boolean indicating whether to checkpoint the learner.
        """

        super().__init__(
            agents=agents,
            agent_types=agent_types,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            target_policy_networks=target_policy_networks,
            target_critic_networks=target_critic_networks,
            discount=discount,
            target_averaging=target_averaging,
            target_update_period=target_update_period,
            target_update_rate=target_update_rate,
            dataset=dataset,
            observation_networks=observation_networks,
            target_observation_networks=target_observation_networks,
            shared_weights=shared_weights,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            max_gradient_norm=max_gradient_norm,
            counter=counter,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
            bootstrap_n=bootstrap_n,
        )


class MAD4PGCentralisedRecurrentTrainer(
    MAD4PGBaseRecurrentTrainer, MADDPGCentralisedRecurrentTrainer
):
    """MAD4PG trainer.
    This is the trainer component of a MAD4PG system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        target_policy_networks: Dict[str, snt.Module],
        target_critic_networks: Dict[str, snt.Module],
        policy_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        critic_optimizer: snt.Optimizer,
        discount: float,
        target_averaging: bool,
        target_update_period: int,
        target_update_rate: float,
        dataset: tf.data.Dataset,
        observation_networks: Dict[str, snt.Module],
        target_observation_networks: Dict[str, snt.Module],
        shared_weights: bool = False,
        max_gradient_norm: float = None,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
        bootstrap_n: int = 10,
    ):
        """Initializes the learner.
        Args:
          policy_network: the online (optimized) policy.
          critic_network: the online critic.
          target_policy_network: the target policy (which lags behind the online
            policy).
          target_critic_network: the target critic.
          discount: discount to use for TD updates.
          target_update_period: number of learner steps to perform before updating
            the target networks.
          dataset: dataset to learn from, whether fixed or from a replay buffer
            (see `acme.datasets.reverb.make_dataset` documentation).
          observation_network: an optional online network to process observations
            before the policy and the critic.
          target_observation_network: the target observation network.
          policy_optimizer: the optimizer to be applied to the DPG (policy) loss.
          critic_optimizer: the optimizer to be applied to the critic loss.
          clipping: whether to clip gradients by global norm.
          counter: counter object used to keep track of steps.
          logger: logger object to be used by learner.
          checkpoint: boolean indicating whether to checkpoint the learner.
        """

        super().__init__(
            agents=agents,
            agent_types=agent_types,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            target_policy_networks=target_policy_networks,
            target_critic_networks=target_critic_networks,
            discount=discount,
            target_averaging=target_averaging,
            target_update_period=target_update_period,
            target_update_rate=target_update_rate,
            dataset=dataset,
            observation_networks=observation_networks,
            target_observation_networks=target_observation_networks,
            shared_weights=shared_weights,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            max_gradient_norm=max_gradient_norm,
            counter=counter,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
            bootstrap_n=bootstrap_n,
        )


class MAD4PGStateBasedRecurrentTrainer(
    MAD4PGBaseRecurrentTrainer, MADDPGStateBasedRecurrentTrainer
):
    """MAD4PG trainer.
    This is the trainer component of a MAD4PG system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        target_policy_networks: Dict[str, snt.Module],
        target_critic_networks: Dict[str, snt.Module],
        policy_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        critic_optimizer: snt.Optimizer,
        discount: float,
        target_averaging: bool,
        target_update_period: int,
        target_update_rate: float,
        dataset: tf.data.Dataset,
        observation_networks: Dict[str, snt.Module],
        target_observation_networks: Dict[str, snt.Module],
        shared_weights: bool = False,
        max_gradient_norm: float = None,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
        bootstrap_n: int = 10,
    ):
        """Initializes the learner.
        Args:
          policy_network: the online (optimized) policy.
          critic_network: the online critic.
          target_policy_network: the target policy (which lags behind the online
            policy).
          target_critic_network: the target critic.
          discount: discount to use for TD updates.
          target_update_period: number of learner steps to perform before updating
            the target networks.
          dataset: dataset to learn from, whether fixed or from a replay buffer
            (see `acme.datasets.reverb.make_dataset` documentation).
          observation_network: an optional online network to process observations
            before the policy and the critic.
          target_observation_network: the target observation network.
          policy_optimizer: the optimizer to be applied to the DPG (policy) loss.
          critic_optimizer: the optimizer to be applied to the critic loss.
          clipping: whether to clip gradients by global norm.
          counter: counter object used to keep track of steps.
          logger: logger object to be used by learner.
          checkpoint: boolean indicating whether to checkpoint the learner.
        """

        super().__init__(
            agents=agents,
            agent_types=agent_types,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            target_policy_networks=target_policy_networks,
            target_critic_networks=target_critic_networks,
            discount=discount,
            target_averaging=target_averaging,
            target_update_period=target_update_period,
            target_update_rate=target_update_rate,
            dataset=dataset,
            observation_networks=observation_networks,
            target_observation_networks=target_observation_networks,
            shared_weights=shared_weights,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            max_gradient_norm=max_gradient_norm,
            counter=counter,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
            bootstrap_n=bootstrap_n,
        )



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic environment loop wrapper to track system statistics"""

import time
from typing import Any, Dict, List, Sequence

import numpy as np
import tensorflow as tf
from acme.utils import loggers

import mava
from mava.utils import training_utils as train_utils
from mava.utils.loggers import Logger
from mava.utils.wrapper_utils import RunningStatistics


class TrainerWrapperBase(mava.Trainer):
    """A base trainer stats class that wrappers a trainer and logs
    certain statistics.
    """

    def __init__(
        self,
        trainer: mava.Trainer,
    ) -> None:
        self._trainer = trainer

    def get_variables(self, names: Sequence[str]) -> Dict[str, Dict[str, np.ndarray]]:
        return self._trainer.get_variables(names)

    def _create_loggers(self, keys: List[str]) -> None:
        raise NotImplementedError

    def __getattr__(self, name: str) -> Any:
        """Expose any other attributes of the underlying trainer."""
        return getattr(self._trainer, name)


class TrainerStatisticsBase(TrainerWrapperBase):
    def __init__(
        self,
        trainer: mava.Trainer,
    ) -> None:
        super().__init__(trainer)
        self._require_loggers = True

    def step(self) -> None:
        # Run the learning step.
        fetches = self._step()

        if self._require_loggers:
            self._create_loggers(list(fetches.keys()))
            self._require_loggers = False

        # compute statistics
        self._compute_statistics(fetches)

        # Compute elapsed time.
        timestamp = time.time()
        elapsed_time = timestamp - self._timestamp if self._timestamp else 0
        self._timestamp: float = timestamp

        # Update our counts and record it.
        counts = self._counter.increment(steps=1, walltime=elapsed_time)
        fetches.update(counts)

        if self._system_checkpointer:
            train_utils.checkpoint_networks(self._system_checkpointer)

        if self._logger:
            self._logger.write(fetches)


class DetailedTrainerStatistics(TrainerStatisticsBase):
    """A trainer class that logs episode stats."""

    def __init__(
        self,
        trainer: mava.Trainer,
        metrics: List[str] = ["policy_loss"],
        summary_stats: List = ["mean", "max", "min", "var", "std"],
    ) -> None:
        super().__init__(trainer)

        self._metrics = metrics
        self._summary_stats = summary_stats

    def _create_loggers(self, keys: List[str]) -> None:

        # get system logger data
        trainer_label = self._logger._label
        base_dir = self._logger._directory
        (
            to_terminal,
            to_csv,
            to_tensorboard,
            time_delta,
            print_fn,
            time_stamp,
        ) = self._logger._logger_info

        self._network_running_statistics: Dict[str, Dict[str, float]] = {}
        self._networks_stats: Dict[str, Dict[str, RunningStatistics]] = {
            key: {} for key in keys
        }
        self._network_loggers: Dict[str, loggers.Logger] = {}

        # statistics dictionary
        for key in keys:
            network_label = trainer_label + "_" + key
            self._network_loggers[key] = Logger(
                label=network_label,
                directory=base_dir,
                to_terminal=to_terminal,
                to_csv=to_csv,
                to_tensorboard=to_tensorboard,
                time_delta=0,
                print_fn=print_fn,
                time_stamp=time_stamp,
            )
            for metric in self._metrics:
                self._networks_stats[key][metric] = RunningStatistics(f"{key}_{metric}")

    def _compute_statistics(self, data: Dict[str, Dict[str, float]]) -> None:
        for network, datum in data.items():
            for key, val in datum.items():
                network_running_statistics: Dict[str, float] = {}
                network_running_statistics[f"{network}_raw_{key}"] = val
                self._networks_stats[network][key].push(val)
                for stat in self._summary_stats:
                    network_running_statistics[
                        f"{network}_{stat}_{key}"
                    ] = self._networks_stats[network][key].__getattribute__(stat)()

                self._network_loggers[network].write(network_running_statistics)


class DetailedTrainerStatisticsWithEpsilon(DetailedTrainerStatistics):
    """Custom DetailedTrainerStatistics class for exposing get_epsilon()"""

    def __init__(
        self,
        trainer: mava.Trainer,
        metrics: List[str] = ["q_value_loss"],
        summary_stats: List = ["mean", "max", "min", "var", "std"],
    ) -> None:
        super().__init__(trainer, metrics, summary_stats)

    def get_epsilon(self) -> float:
        return self._trainer.get_epsilon()  # type: ignore

    def get_trainer_steps(self) -> float:
        return self._trainer.get_trainer_steps()  # type: ignore

    def step(self) -> None:
        # Run the learning step.
        fetches = self._step()

        if self._require_loggers:
            self._create_loggers(list(fetches.keys()))
            self._require_loggers = False

        # compute statistics
        self._compute_statistics(fetches)

        timestamp = time.time()
        elapsed_time = timestamp - self._timestamp if self._timestamp else 0
        self._timestamp = timestamp

        # Update our counts and record it.
        counts = self._counter.increment(steps=1, walltime=elapsed_time)
        fetches.update(counts)

        if self._system_checkpointer:
            train_utils.checkpoint_networks(self._system_checkpointer)

        fetches["epsilon"] = self.get_epsilon()
        self._trainer._decrement_epsilon()  # type: ignore

        if self._logger:
            self._logger.write(fetches)


# TODO(Kale-ab): Is there a better way to do this?
# Maybe using hooks or callbacks.
class NetworkStatisticsBase(TrainerWrapperBase):
    """
    A base class for logging network statistics.
        gradient_norms: List of norms (see tf.norm.ord) to apply to grads.
        weight_norms: List of norms (see tf.norm.ord) to apply to weights.
        log_interval: Log every [log_interval] learner steps.
    """

    def __init__(
        self,
        trainer: mava.Trainer,
        # Log only l2 norm by default.
        gradient_norms: List,
        weight_norms: List,
        log_interval: int,
        log_weights: bool,
        log_gradients: bool,
    ) -> None:
        super().__init__(trainer)

        self._network_loggers: Dict[str, loggers.Logger] = {}
        self.gradient_norms = gradient_norms
        self.weight_norms = weight_norms
        self._create_loggers(self._agents)
        self.log_interval = log_interval
        self.log_weights = log_weights
        self.log_gradients = log_gradients

        assert (
            self.log_weights or self.log_gradients
        ), "Nothing is selected to be logged."

    def _log_step(self) -> bool:
        return bool(
            self._counter
            and self._counter._counts
            and self._counter._counts.get("steps")
            and self._counter._counts.get("steps") % self.log_interval == 0
        )

    def _create_loggers(self, keys: List[str]) -> None:
        trainer_label = self._logger._label
        base_dir = self._logger._directory
        (
            to_terminal,
            to_csv,
            to_tensorboard,
            time_delta,
            print_fn,
            time_stamp,
        ) = self._logger._logger_info

        trainer_label = f"{trainer_label}_networks_stats"
        for key in keys:
            network_label = trainer_label + "_" + key
            self._network_loggers[key] = Logger(
                label=network_label,
                directory=base_dir,
                to_terminal=False,
                to_csv=False,
                to_tensorboard=to_tensorboard,
                print_fn=print_fn,
                time_stamp=time_stamp,
                time_delta=0,
            )

    # Function determines which weight layers to log.
    # We are usually only concerned with weights and grads of linear and conv layers.
    # TODO(Kale-ab) Can we be more robust.
    # Try getting layer type from policy or critic networks
    def _log_data(self, name: str) -> bool:
        # Log linear and conv weights and not bias units.
        return ("linear" in name.lower() or "conv" in name.lower()) and not (
            "b:" in name.lower()
        )

    def _apply_norms(self, value: tf.Tensor, norms_list: List) -> Dict:
        return {norm: tf.norm(value, ord=norm).numpy() for norm in norms_list}

    def _log_gradients(
        self, label: str, agent: str, variables_names: List, gradients: List
    ) -> None:
        assert len(variables_names) == len(
            gradients
        ), "Variable names and gradients do not match"
        grads_dict = {}
        # Log Grads per layer
        for index, grad in enumerate(gradients):
            variables_name = variables_names[index]
            if self._log_data(variables_name):
                grads_dict[f"{label}/grad/{variables_name}"] = grad
                grads_dict[f"{label}/gradnorm/{variables_name}"] = self._apply_norms(
                    grad, self.gradient_norms
                )

        # Log whole network grads
        all_grads_flat = tf.concat([tf.reshape(grad, -1) for grad in gradients], axis=0)
        grads_dict[f"{label}/grad/wholenetwork"] = all_grads_flat
        grads_dict[f"{label}/gradnorm/wholenetwork"] = self._apply_norms(
            all_grads_flat, self.gradient_norms
        )

        self._network_loggers[agent].write(grads_dict)

    def _log_weights(self, label: str, agent: str, weights: List) -> None:
        weights_dict = {}

        # Log Weights per layer
        for weight in weights:
            if self._log_data(weight.name):
                weights_dict[f"{label}/weight/{weight.name}"] = weight
                weights_dict[f"{label}/weightnorm/{weight.name}"] = self._apply_norms(
                    weight, self.weight_norms
                )

        # Log whole network weights
        all_weights_flat = tf.concat(
            [tf.reshape(weight, -1) for weight in weights], axis=0
        )
        weights_dict[f"{label}/weight/wholenetwork"] = all_weights_flat
        weights_dict[f"{label}/weightnorm/wholenetwork"] = self._apply_norms(
            all_weights_flat, self.weight_norms
        )

        self._network_loggers[agent].write(weights_dict)

    def step(self) -> None:
        fetches = self._step()

        # Compute elapsed time.
        timestamp = time.time()
        elapsed_time = timestamp - self._timestamp if self._timestamp else 0
        self._timestamp: float = timestamp

        # Update our counts and record it.
        counts = self._counter.increment(steps=1, walltime=elapsed_time)
        fetches.update(counts)

        if self._system_checkpointer:
            train_utils.checkpoint_networks(self._system_checkpointer)

        if self._logger:
            self._logger.write(fetches)


class NetworkStatistics(NetworkStatisticsBase):
    """
    A class for logging network statistics.
    This class assumes the trainer has the following:
        _forward: Forward pass. Stores a policy loss and tf.GradientTape.
        _backward: Updates network using policy loss and tf.GradientTape.
    """

    def __init__(
        self,
        trainer: mava.Trainer,
        # Log only l2 norm by default.
        gradient_norms: List = [2],
        weight_norms: List = [2],
        log_interval: int = 100,
        log_weights: bool = True,
        log_gradients: bool = True,
    ) -> None:
        super().__init__(
            trainer,
            gradient_norms,
            weight_norms,
            log_interval,
            log_weights,
            log_gradients,
        )

    def _step(
        self,
    ) -> Dict[str, Dict[str, Any]]:

        # Update the target networks
        # Trying not to assume off policy.
        if hasattr(self, "_update_target_networks"):
            self._update_target_networks()

        # Get data from replay (dropping extras if any). Note there is no
        # extra data here because we do not insert any into Reverb.
        inputs = next(self._iterator)

        self._forward(inputs)

        self._backward()

        # Log losses per agent
        return self.policy_losses

    def _backward(self) -> None:
        policy_losses = self.policy_losses
        tape = self.tape
        log_current_timestep = self._log_step()

        # Calculate the gradients and update the networks
        for agent in self._agents:
            agent_key = self.agent_net_keys[agent]

            # Get trainable variables.
            policy_variables = (
                self._observation_networks[agent_key].trainable_variables
                + self._policy_networks[agent_key].trainable_variables
            )

            # Compute gradients.
            policy_gradients = tape.gradient(policy_losses[agent], policy_variables)

            # Maybe clip gradients.
            if self._clipping:
                policy_gradients = tf.clip_by_global_norm(policy_gradients, 40.0)[0]

            # Apply gradients.
            self._policy_optimizers[agent_key].apply(policy_gradients, policy_variables)

            if log_current_timestep:
                if self.log_weights:
                    self._log_weights(
                        label="Policy", agent=agent, weights=policy_variables
                    )
                if self.log_gradients:
                    self._log_gradients(
                        label="Policy",
                        agent=agent,
                        variables_names=[vars.name for vars in policy_variables],
                        gradients=policy_gradients,
                    )
        train_utils.safe_del(self, "tape")


class NetworkStatisticsActorCritic(NetworkStatisticsBase):
    """
    A class for logging network statistics.
    This class assumes the trainer has the following:
        _forward: Forward pass. Stores a policy loss, critic loss and tf.GradientTape.
        _backward: Updates network using policy loss, critic loss and tf.GradientTape.
    """

    def __init__(
        self,
        trainer: mava.Trainer,
        # Log only l2 norm by default.
        gradient_norms: List = [2],
        weight_norms: List = [2],
        log_interval: int = 100,
        log_weights: bool = True,
        log_gradients: bool = True,
    ) -> None:
        super().__init__(
            trainer,
            gradient_norms,
            weight_norms,
            log_interval,
            log_weights,
            log_gradients,
        )

    def _step(
        self,
    ) -> Dict[str, Dict[str, Any]]:

        # Update the target networks
        # Trying not to assume off policy.
        if hasattr(self, "_update_target_networks"):
            self._update_target_networks()

        # Get data from replay (dropping extras if any). Note there is no
        # extra data here because we do not insert any into Reverb.
        inputs = next(self._iterator)

        self._forward(inputs)

        self._backward()

        # Log losses per agent
        return train_utils.map_losses_per_agent_ac(
            self.critic_losses, self.policy_losses
        )

    def _backward(self) -> None:
        # Calculate the gradients and update the networks
        policy_losses = self.policy_losses
        critic_losses = self.critic_losses
        tape = self.tape
        log_current_timestep = self._log_step()

        for agent in self._agents:
            agent_key = self.agent_net_keys[agent]

            # Get trainable variables.
            policy_variables = (
                self._observation_networks[agent_key].trainable_variables
                + self._policy_networks[agent_key].trainable_variables
            )
            critic_variables = (
                # In this agent, the critic loss trains the observation network.
                self._observation_networks[agent_key].trainable_variables
                + self._critic_networks[agent_key].trainable_variables
            )

            # Compute gradients.
            policy_gradients = tape.gradient(policy_losses[agent], policy_variables)
            critic_gradients = tape.gradient(critic_losses[agent], critic_variables)

            # Maybe clip gradients.
            policy_gradients = tf.clip_by_global_norm(
                policy_gradients, self._max_gradient_norm
            )[0]
            critic_gradients = tf.clip_by_global_norm(
                critic_gradients, self._max_gradient_norm
            )[0]

            # Apply gradients.
            self._policy_optimizers[agent_key].apply(policy_gradients, policy_variables)
            self._critic_optimizers[agent_key].apply(critic_gradients, critic_variables)

            if log_current_timestep:
                if self.log_weights:
                    self._log_weights(
                        label="Policy", agent=agent, weights=policy_variables
                    )
                    self._log_weights(
                        label="Critic", agent=agent, weights=critic_variables
                    )

                if self.log_gradients:
                    self._log_gradients(
                        label="Policy",
                        agent=agent,
                        variables_names=[vars.name for vars in policy_variables],
                        gradients=policy_gradients,
                    )
                    self._log_gradients(
                        label="Critic",
                        agent=agent,
                        variables_names=[vars.name for vars in critic_variables],
                        gradients=critic_gradients,
                    )

        train_utils.safe_del(self, "tape")



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import gym
import numpy as np
from pettingzoo.utils.wrappers import OrderEnforcingWrapper as PettingzooWrapper
from supersuit.aec_wrappers import ObservationWrapper as SequentialObservationWrapper
from supersuit.parallel_wrappers import ObservationWrapper as ParallelObservationWrapper
from supersuit.parallel_wrappers import ParallelWraper as ParallelEnvPettingZoo

from mava.types import Observation, Reward
from mava.utils.wrapper_utils import RunningMeanStd
from mava.wrappers.env_wrappers import ParallelEnvWrapper, SequentialEnvWrapper

# Prevent circular import issue.
if TYPE_CHECKING:
    from mava.wrappers.pettingzoo import (
        PettingZooAECEnvWrapper,
        PettingZooParallelEnvWrapper,
    )

PettingZooEnv = Union["PettingZooAECEnvWrapper", "PettingZooParallelEnvWrapper"]

# TODO(Kale-ab): Make wrapper more general
# Should Works across any SequentialEnvWrapper or ParallelEnvWrapper.
"""
GYM Preprocess Wrappers.

Other gym preprocess wrappers:
    https://github.com/PettingZoo-Team/SuperSuit/blob/1f02289e8f51082aa50a413b34700b67042410c6/supersuit/gym_wrappers.py
    https://github.com/openai/gym/tree/master/gym/wrappers
"""


class StandardizeObservationGym(gym.ObservationWrapper):
    """
    Standardize observations
    Ensures mean of 0 and standard deviation of 1 (unit variance) for obs.
    From https://github.com/ikostrikov/pytorch-a3c/blob/e898f7514a03de73a2bf01e7b0f17a6f93963389/envs.py # noqa: E501
    """

    def __init__(self, env: gym.Env = None) -> None:
        super(StandardizeObservationGym, self).__init__(env)
        self.state_mean = 0
        self.state_std = 0
        self.alpha = 0.9999
        self.num_steps = 0

    def _observation(self, observation: Any) -> Any:
        self.num_steps += 1
        self.state_mean = self.state_mean * self.alpha + observation.mean() * (
            1 - self.alpha
        )
        self.state_std = self.state_std * self.alpha + observation.std() * (
            1 - self.alpha
        )

        unbiased_mean = self.state_mean / (1 - pow(self.alpha, self.num_steps))
        unbiased_std = self.state_std / (1 - pow(self.alpha, self.num_steps))

        return (observation - unbiased_mean) / (unbiased_std + 1e-8)


"""
Petting Zoo Env Preprocess Wrappers.

Other PZ preprocess wrappers:
    https://github.com/PettingZoo-Team/SuperSuit
"""


# TODO(Kale-ab): Configure this to use RunningMeanStd.
class StandardizeObservation:
    """
    Standardize observations
    Ensures mean of 0 and standard deviation of 1 (unit variance) for obs.
    Adapted from https://github.com/ikostrikov/pytorch-a3c/blob/e898f7514a03de73a2bf01e7b0f17a6f93963389/envs.py # noqa: E501

        :env Env to wrap.
        :load_params Params to load.
        :alpha To avoid division by zero.
    """

    def __init__(
        self,
        env: PettingZooEnv = None,
        load_params: Dict = None,
        alpha: float = 0.999,
    ):
        self.env: Optional[PettingZooEnv] = env
        self.alpha = alpha
        self._ini_params(load_params)

    def _ini_params(self, load_params: Dict = None) -> None:
        if load_params:
            self._internal_state = load_params
        else:
            params = {
                agent: {
                    "state_mean": 0,
                    "state_std": 0,
                    "num_steps": 0,
                }
                for agent in self.env.possible_agents  # type:ignore
            }
            self._internal_state = params

    def _get_updated_observation(
        self,
        agent: str,
        observation: Observation,
    ) -> Observation:

        state_mean = self._internal_state[agent].get("state_mean")
        state_std = self._internal_state[agent].get("state_std")
        num_steps = self._internal_state[agent].get("num_steps")

        state_mean = state_mean * self.alpha + observation.mean() * (  # type:ignore
            1 - self.alpha
        )
        state_std = state_std * self.alpha + observation.std() * (  # type:ignore
            1 - self.alpha
        )

        unbiased_mean = state_mean / (1 - pow(self.alpha, num_steps))
        unbiased_std = state_std / (1 - pow(self.alpha, num_steps))

        self._internal_state[agent]["state_mean"] = state_mean
        self._internal_state[agent]["state_std"] = state_std

        return (observation - unbiased_mean) / (unbiased_std + 1e-8)


class StandardizeObservationSequential(
    SequentialObservationWrapper, StandardizeObservation
):
    """Standardize Obs in Sequential Env"""

    def __init__(
        self, env: PettingZooEnv, load_params: Dict = None, alpha: float = 0.999
    ) -> None:
        SequentialObservationWrapper.__init__(self, env)
        StandardizeObservation.__init__(self, env, load_params, alpha)

    def _modify_observation(self, agent: str, observation: Observation) -> Observation:
        self._internal_state[agent]["num_steps"] = (
            int(self._internal_state[agent]["num_steps"]) + 1
        )
        return self._get_updated_observation(agent, observation)

    def _check_wrapper_params(self) -> None:
        return

    def _modify_spaces(self) -> None:
        return


class StandardizeObservationParallel(
    ParallelObservationWrapper, StandardizeObservation
):
    """Standardize Obs in Parallel Env"""

    def __init__(
        self, env: PettingZooEnv, load_params: Dict = None, alpha: float = 0.999
    ) -> None:
        ParallelObservationWrapper.__init__(self, env)
        StandardizeObservation.__init__(self, env, load_params, alpha)

    def _modify_observation(self, agent: str, observation: Observation) -> Observation:
        self._internal_state[agent]["num_steps"] = (
            int(self._internal_state[agent]["num_steps"]) + 1
        )
        return self._get_updated_observation(agent, observation)

    def _check_wrapper_params(self) -> None:
        return

    def _modify_spaces(self) -> None:
        return


class StandardizeReward:
    """
    Standardize rewards
    We rescale rewards, but don't shift the mean - http://joschu.net/docs/nuts-and-bolts.pdf .
    Adapted from https://github.com/DLR-RM/stable-baselines3/blob/237223f834fe9b8143ea24235d087c4e32addd2f/stable_baselines3/common/vec_env/vec_normalize.py # noqa: E501
    and https://github.com/openai/baselines/blob/master/baselines/common/vec_env/vec_normalize.py.
        :env Env to wrap.
        :load_params Params to load.
        :upper_bound: Max value for discounted reward.
        :lower_bound: Min value for discounted reward.
        :alpha To avoid division by zero.
    """

    def __init__(
        self,
        env: PettingZooEnv = None,
        load_params: Dict = None,
        lower_bound: float = -10.0,
        upper_bound: float = 10.0,
        alpha: float = 0.999,
    ):
        self._ini_params(load_params)
        self.env = env
        self.upper_bound = upper_bound
        self.lower_bound = lower_bound
        self.alpha = alpha

    def _ini_params(self, load_params: Dict = None) -> None:
        if load_params:
            self._internal_state = load_params
        else:
            params = {}
            gamma_dict = None
            if hasattr(self, "_discounts"):
                gamma_dict = self._discount  # type: ignore

            for agent in self.env.possible_agents:  # type:ignore
                gamma = gamma_dict[agent] if gamma_dict else 1
                params[agent] = {
                    "return": 0,
                    "ret_rms": RunningMeanStd(shape=()),
                    "gamma": gamma,
                }
            self._internal_state = params

    def _update_reward(self, agent: str, reward: Reward) -> None:
        """Update reward normalization statistics."""
        agent_dict = self._internal_state[agent]
        ret = agent_dict["return"]
        gamma = agent_dict["gamma"]
        ret_rms = agent_dict["ret_rms"]

        ret = ret * gamma + reward
        ret_rms.update(ret)

        agent_dict["return"] = ret
        agent_dict["ret_rms"] = ret_rms
        self._internal_state[agent] = agent_dict

    def normalize_reward(self, agent: str, reward: Reward) -> Reward:
        """
        Normalize rewards using rewards statistics.
        Calling this method does not update statistics.
        """
        agent_dict = self._internal_state[agent]
        ret_rms = agent_dict["ret_rms"]
        reward = np.clip(
            reward / np.sqrt(ret_rms.var + self.alpha),
            self.lower_bound,
            self.upper_bound,
        )
        return reward

    def _get_updated_reward(
        self,
        agent: str,
        reward: Reward,
    ) -> Reward:
        self._update_reward(agent, reward)
        reward = self.normalize_reward(agent, reward)
        return reward


class StandardizeRewardSequential(PettingzooWrapper, StandardizeReward):
    def __init__(
        self,
        env: SequentialEnvWrapper,
        load_params: Dict = None,
        lower_bound: float = -10.0,
        upper_bound: float = 10.0,
        alpha: float = 0.999,
    ) -> None:
        PettingzooWrapper.__init__(self, env)
        StandardizeReward.__init__(
            self, env, load_params, lower_bound, upper_bound, alpha
        )

    def reset(self) -> None:
        # Reset returns, but not running scores.
        for stats in self._internal_state.values():
            stats["return"] = 0

        super().reset()
        self.rewards = {
            agent: self._get_updated_reward(agent, reward)
            for agent, reward in self.rewards.items()  # type: ignore
        }
        self.__cumulative_rewards = {a: 0 for a in self.agents}
        self._accumulate_rewards()

    def step(self, action: np.ndarray) -> None:
        agent = self.env.agent_selection  # type: ignore
        super().step(action)
        self.rewards = {
            agent: self._get_updated_reward(agent, reward)
            for agent, reward in self.rewards.items()
        }
        self.__cumulative_rewards[agent] = 0
        self._cumulative_rewards = self.__cumulative_rewards
        self._accumulate_rewards()


class StandardizeRewardParallel(
    ParallelEnvPettingZoo,
    StandardizeReward,
):
    def __init__(
        self,
        env: ParallelEnvWrapper,
        load_params: Dict = None,
        lower_bound: float = -10.0,
        upper_bound: float = 10.0,
        alpha: float = 0.999,
    ) -> None:
        ParallelEnvPettingZoo.__init__(self, env)
        StandardizeReward.__init__(
            self, env, load_params, lower_bound, upper_bound, alpha
        )

    def reset(self) -> Observation:
        # Reset returns, but not running scores.
        for stats in self._internal_state.values():
            stats["return"] = 0

        obs = self.env.reset()  # type: ignore
        self.agents = self.env.agents  # type: ignore
        return obs

    def step(self, actions: Dict) -> Any:
        obs, rew, done, info = super().step(actions)
        rew = {
            agent: self._get_updated_reward(agent, rew) for agent, rew in rew.items()
        }
        return obs, rew, done, info



# Adapted from https://github.com/openai/multiagent-particle-envs.

import importlib
from typing import Any


def load(name: str) -> Any:
    return importlib.import_module(f"mava.utils.debugging.scenarios.{name}")



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore

import os
import pickle

from acme.tf import savers as tf2_savers


def save_policy_and_obs(architecture, agent_type, checkpoint_subpath, pickle_save_loc):
    policy_networks = architecture.create_actor_variables()
    before_sum = policy_networks["policies"][agent_type].variables[1].numpy().sum()
    objects_to_save = {
        "policy": policy_networks["policies"][agent_type],
        "observation": policy_networks["observations"][agent_type],
    }

    checkpointer_dir = os.path.join(checkpoint_subpath, agent_type)
    tf2_savers.Checkpointer(
        time_delta_minutes=1,
        add_uid=False,
        directory=checkpointer_dir,
        objects_to_save=objects_to_save,
        enable_checkpointing=True,
    )
    after_sum = policy_networks["policies"][agent_type].variables[1].numpy().sum()

    assert before_sum != after_sum

    # Save policy variables
    policy = policy_networks["policies"][agent_type].set.variables
    policy_file = open(os.path.join(pickle_save_loc, "policy.obj"), "wb")
    pickle.dump(policy, policy_file)
    policy_file.close()

    # So observation variables
    observation = policy_networks["observations"][agent_type].variables
    obs_file = open(os.path.join(pickle_save_loc, "observations.obj"), "wb")
    pickle.dump(observation, obs_file)
    obs_file.close()


def load_policy_and_obs(pickle_save_loc):
    # Load policy variables
    filehandler = open(os.path.join(pickle_save_loc, "policy.obj"), "rb")
    policy_net = pickle.load(filehandler)

    # Load observation variables
    filehandler = open(os.path.join(pickle_save_loc, "policy.obj"), "rb")
    obs_net = pickle.load(filehandler)

    return policy_net, obs_net



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Executor implementations for MADQN."""

from typing import Any, Dict, Optional

import dm_env
import numpy as np
import sonnet as snt
import tensorflow as tf
from acme import types

# Internal imports.
from acme.tf import utils as tf2_utils
from acme.tf import variable_utils as tf2_variable_utils

from mava import adders
from mava.components.tf.modules.communication import BaseCommunicationModule
from mava.systems.tf.executors import (
    FeedForwardExecutor,
    RecurrentCommExecutor,
    RecurrentExecutor,
)
from mava.systems.tf.madqn.training import MADQNTrainer
from mava.types import OLT


class MADQNFeedForwardExecutor(FeedForwardExecutor):
    """A feed-forward executor.
    An executor based on a feed-forward policy for each agent in the system
    which takes non-batched observations and outputs non-batched actions.
    It also allows adding experiences to replay and updating the weights
    from the policy on the learner.
    """

    def __init__(
        self,
        q_networks: Dict[str, snt.Module],
        action_selectors: Dict[str, snt.Module],
        trainer: MADQNTrainer,
        shared_weights: bool = True,
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
        communication_module: Optional[BaseCommunicationModule] = None,
        fingerprint: bool = False,
        evaluator: bool = False,
    ):
        """Initializes the executor.
        Args:
          policy_network: the policy to run for each agent in the system.
          shared_weights: specify if weights are shared between agent networks.
          adder: the adder object to which allows to add experiences to a
            dataset/replay buffer.
          variable_client: object which allows to copy weights from the trainer copy
            of the policies to the executor copy (in case they are separate).
        """

        # Store these for later use.
        self._adder = adder
        self._variable_client = variable_client
        self._q_networks = q_networks
        self._action_selectors = action_selectors
        self._trainer = trainer
        self._shared_weights = shared_weights
        self._fingerprint = fingerprint
        self._evaluator = evaluator

    @tf.function
    def _policy(
        self,
        agent: str,
        observation: types.NestedTensor,
        legal_actions: types.NestedTensor,
        epsilon: tf.Tensor,
        fingerprint: Optional[tf.Tensor] = None,
    ) -> types.NestedTensor:

        # Add a dummy batch dimension and as a side effect convert numpy to TF.
        batched_observation = tf2_utils.add_batch_dim(observation)
        batched_legals = tf2_utils.add_batch_dim(legal_actions)

        # index network either on agent type or on agent id
        agent_key = agent.split("_")[0] if self._shared_weights else agent

        # Compute the policy, conditioned on the observation and
        # possibly the fingerprint.
        if fingerprint is not None:
            q_values = self._q_networks[agent_key](batched_observation, fingerprint)
        else:
            q_values = self._q_networks[agent_key](batched_observation)

        # select legal action
        action = self._action_selectors[agent_key](
            q_values, batched_legals, epsilon=epsilon
        )

        return action

    def select_action(
        self, agent: str, observation: types.NestedArray
    ) -> types.NestedArray:
        raise NotImplementedError

    def observe_first(
        self,
        timestep: dm_env.TimeStep,
        extras: Dict[str, types.NestedArray] = {},
    ) -> None:

        if self._fingerprint and self._trainer is not None:
            epsilon = self._trainer.get_epsilon()
            trainer_step = self._trainer.get_trainer_steps()
            fingerprint = np.array([epsilon, trainer_step])
            extras.update({"fingerprint": fingerprint})

        if self._adder:
            self._adder.add_first(timestep, extras)

    def observe(
        self,
        actions: Dict[str, types.NestedArray],
        next_timestep: dm_env.TimeStep,
        next_extras: Dict[str, types.NestedArray] = {},
    ) -> None:
        if self._fingerprint and self._trainer is not None:
            trainer_step = self._trainer.get_trainer_steps()
            epsilon = self._trainer.get_epsilon()
            fingerprint = np.array([epsilon, trainer_step])
            next_extras.update({"fingerprint": fingerprint})

        if self._adder:
            self._adder.add(actions, next_timestep, next_extras)

    def select_actions(
        self, observations: Dict[str, OLT]
    ) -> Dict[str, types.NestedArray]:
        actions = {}
        for agent, observation in observations.items():
            # Pass the observation through the policy network.
            if not self._evaluator:
                epsilon = self._trainer.get_epsilon()
            else:
                # Note (dries): For some reason 0 epsilon breaks on StarCraft.
                epsilon = 1e-10

            epsilon = tf.convert_to_tensor(epsilon)

            if self._fingerprint:
                trainer_step = self._trainer.get_trainer_steps()
                fingerprint = tf.concat([epsilon, trainer_step], axis=0)
                fingerprint = tf.expand_dims(fingerprint, axis=0)
                fingerprint = tf.cast(fingerprint, "float32")
            else:
                fingerprint = None

            action = self._policy(
                agent,
                observation.observation,
                observation.legal_actions,
                epsilon,
                fingerprint,
            )

            actions[agent] = tf2_utils.to_numpy_squeeze(action)

        # Return a numpy array with squeezed out batch dimension.
        return actions

    def update(self, wait: bool = False) -> None:
        if self._variable_client:
            self._variable_client.update(wait)


class MADQNRecurrentExecutor(RecurrentExecutor):
    """A feed-forward executor.
    An executor based on a feed-forward policy for each agent in the system
    which takes non-batched observations and outputs non-batched actions.
    It also allows adding experiences to replay and updating the weights
    from the policy on the learner.
    """

    def __init__(
        self,
        q_networks: Dict[str, snt.Module],
        action_selectors: Dict[str, snt.Module],
        shared_weights: bool = True,
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
        store_recurrent_state: bool = True,
        trainer: MADQNTrainer = None,
        communication_module: Optional[BaseCommunicationModule] = None,
        evaluator: bool = False,
        fingerprint: bool = False,
    ):
        """Initializes the executor.
        Args:
          policy_network: the policy to run for each agent in the system.
          shared_weights: specify if weights are shared between agent networks.
          adder: the adder object to which allows to add experiences to a
            dataset/replay buffer.
          variable_client: object which allows to copy weights from the trainer copy
            of the policies to the executor copy (in case they are separate).
        """

        # Store these for later use.
        self._adder = adder
        self._variable_client = variable_client
        self._q_networks = q_networks
        self._policy_networks = q_networks
        self._action_selectors = action_selectors
        self._store_recurrent_state = store_recurrent_state
        self._trainer = trainer
        self._shared_weights = shared_weights

        self._states: Dict[str, Any] = {}

    @tf.function
    def _policy(
        self,
        agent: str,
        observation: types.NestedTensor,
        state: types.NestedTensor,
        legal_actions: types.NestedTensor,
        epsilon: tf.Tensor,
    ) -> types.NestedTensor:

        # Add a dummy batch dimension and as a side effect convert numpy to TF.
        batched_observation = tf2_utils.add_batch_dim(observation)
        batched_legals = tf2_utils.add_batch_dim(legal_actions)

        # index network either on agent type or on agent id
        agent_key = agent.split("_")[0] if self._shared_weights else agent

        # Compute the policy, conditioned on the observation.
        q_values, new_state = self._q_networks[agent_key](batched_observation, state)

        # select legal action
        action = self._action_selectors[agent_key](
            q_values, batched_legals, epsilon=epsilon
        )

        return action, new_state

    def select_action(
        self, agent: str, observation: types.NestedArray
    ) -> types.NestedArray:
        raise NotImplementedError

    def select_actions(
        self, observations: Dict[str, OLT]
    ) -> Dict[str, types.NestedArray]:
        actions = {}

        for agent, observation in observations.items():

            # Pass the observation through the policy network.
            if self._trainer is not None:
                epsilon = self._trainer.get_epsilon()
            else:
                epsilon = 0.0

            epsilon = tf.convert_to_tensor(epsilon)

            policy_output, new_state = self._policy(
                agent,
                observation.observation,
                self._states[agent],
                observation.legal_actions,
                epsilon,
            )

            self._states[agent] = new_state

            actions[agent] = tf2_utils.to_numpy_squeeze(policy_output)

        # Return a numpy array with squeezed out batch dimension.
        return actions


class MADQNRecurrentCommExecutor(RecurrentCommExecutor):
    """A feed-forward executor.
    An executor based on a feed-forward policy for each agent in the system
    which takes non-batched observations and outputs non-batched actions.
    It also allows adding experiences to replay and updating the weights
    from the policy on the learner.
    """

    def __init__(
        self,
        q_networks: Dict[str, snt.Module],
        action_selectors: Dict[str, snt.Module],
        communication_module: BaseCommunicationModule,
        shared_weights: bool = True,
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
        store_recurrent_state: bool = True,
        trainer: MADQNTrainer = None,
        fingerprint: bool = False,
        evaluator: bool = False,
    ):
        """Initializes the executor.
        Args:
          policy_network: the policy to run for each agent in the system.
          shared_weights: specify if weights are shared between agent networks.
          adder: the adder object to which allows to add experiences to a
            dataset/replay buffer.
          variable_client: object which allows to copy weights from the trainer copy
            of the policies to the executor copy (in case they are separate).
        """

        # Store these for later use.
        self._adder = adder
        self._variable_client = variable_client
        self._q_networks = q_networks
        self._policy_networks = q_networks
        self._communication_module = communication_module
        self._action_selectors = action_selectors
        self._store_recurrent_state = store_recurrent_state
        self._trainer = trainer
        self._shared_weights = shared_weights

        self._states: Dict[str, Any] = {}
        self._messages: Dict[str, Any] = {}

    @tf.function
    def _policy(
        self,
        agent: str,
        observation: types.NestedTensor,
        state: types.NestedTensor,
        message: types.NestedTensor,
        legal_actions: types.NestedTensor,
        epsilon: tf.Tensor,
    ) -> types.NestedTensor:

        # Add a dummy batch dimension and as a side effect convert numpy to TF.
        batched_observation = tf2_utils.add_batch_dim(observation)
        batched_legals = tf2_utils.add_batch_dim(legal_actions)

        # index network either on agent type or on agent id
        agent_key = agent.split("_")[0] if self._shared_weights else agent

        # Compute the policy, conditioned on the observation.
        (q_values, m_values), new_state = self._q_networks[agent_key](
            batched_observation, state, message
        )

        # select legal action
        action = self._action_selectors[agent_key](
            q_values, batched_legals, epsilon=epsilon
        )

        return (action, m_values), new_state

    def select_action(
        self, agent: str, observation: types.NestedArray
    ) -> types.NestedArray:
        raise NotImplementedError

    def select_actions(
        self, observations: Dict[str, OLT]
    ) -> Dict[str, types.NestedArray]:
        actions = {}

        message_inputs = self._communication_module.process_messages(self._messages)

        for agent, observation in observations.items():

            # Pass the observation through the policy network.
            if self._trainer is not None:
                epsilon = self._trainer.get_epsilon()
            else:
                epsilon = 0.0

            epsilon = tf.convert_to_tensor(epsilon)

            (policy_output, new_message), new_state = self._policy(
                agent,
                observation.observation,
                self._states[agent],
                message_inputs[agent],
                observation.legal_actions,
                epsilon,
            )

            self._states[agent] = new_state
            self._messages[agent] = new_message

            actions[agent] = tf2_utils.to_numpy_squeeze(policy_output)

        # Return a numpy array with squeezed out batch dimension.
        return actions



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional

import sonnet as snt
from acme.specs import EnvironmentSpec
from acme.tf import variable_utils as tf2_variable_utils

from mava import adders
from mava.systems.tf.maddpg.execution import (
    MADDPGFeedForwardExecutor,
    MADDPGRecurrentExecutor,
)


class MAD4PGFeedForwardExecutor(MADDPGFeedForwardExecutor):
    """A feed-forward executor for MAD4PG.
    An executor based on a feed-forward policy for each agent in the system
    which takes non-batched observations and outputs non-batched actions.
    It also allows adding experiences to replay and updating the weights
    from the policy on the learner.
    """

    def __init__(
        self,
        policy_networks: Dict[str, snt.Module],
        agent_specs: Dict[str, EnvironmentSpec],
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
        shared_weights: bool = True,
    ):
        super().__init__(
            policy_networks=policy_networks,
            agent_specs=agent_specs,
            adder=adder,
            variable_client=variable_client,
            shared_weights=shared_weights,
        )


class MAD4PGRecurrentExecutor(MADDPGRecurrentExecutor):
    """A recurrent executor for MAD4PG.
    An executor based on a recurrent policy for each agent in the system
    which takes non-batched observations and outputs non-batched actions.
    It also allows adding experiences to replay and updating the weights
    from the policy on the learner.
    """

    def __init__(
        self,
        policy_networks: Dict[str, snt.Module],
        agent_specs: Dict[str, EnvironmentSpec],
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
        shared_weights: bool = True,
    ):
        super().__init__(
            policy_networks=policy_networks,
            agent_specs=agent_specs,
            adder=adder,
            variable_client=variable_client,
            shared_weights=shared_weights,
        )



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore

from typing import Dict, Optional

import dm_env
from acme import types

from mava.utils.environments.RoboCup_env.robocup_utils.game_object import Flag

true_flag_coords = Flag.FLAG_COORDS


class CustomExecutor:
    """A fixed executor to test in the environments are working."""

    def __init__(self, agents):
        # Convert action and observation specs.
        self.agents = agents

    def observe_first(
        self,
        timestep: dm_env.TimeStep,
        extras: Dict[str, types.NestedArray] = {"": ()},
    ) -> None:
        pass

    def observe(
        self,
        actions: Dict[str, types.NestedArray],
        next_timestep: dm_env.TimeStep,
        next_extras: Optional[Dict[str, types.NestedArray]] = {},
    ) -> None:
        pass

    def select_actions(
        self, observations: Dict[str, types.NestedArray]
    ) -> Dict[str, types.NestedArray]:
        actions = {}
        for agent, observation in observations.items():
            # Pass the observation through the policy network.
            actions[agent] = self.agents[agent].get_action(observation.observation)

        # Return a numpy array with squeezed out batch dimension.
        return actions

    def update(self, wait: bool = False) -> None:
        pass



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MAPPO trainer implementation."""
import copy
import os
import time
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
import sonnet as snt
import tensorflow as tf
import tensorflow_probability as tfp
import trfl
from acme.tf import utils as tf2_utils
from acme.utils import counting, loggers

import mava
from mava.systems.tf import savers as tf2_savers
from mava.utils import training_utils as train_utils

train_utils.set_growing_gpu_memory()

tfd = tfp.distributions


class MAPPOTrainer(mava.Trainer):
    """MAPPO trainer.
    This is the trainer component of a MAPPO system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[Any],
        agent_types: List[str],
        observation_networks: Dict[str, snt.Module],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        dataset: tf.data.Dataset,
        policy_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        critic_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        shared_weights: bool,
        discount: float = 0.99,
        lambda_gae: float = 1.0,
        entropy_cost: float = 0.0,
        baseline_cost: float = 1.0,
        clipping_epsilon: float = 0.2,
        max_gradient_norm: Optional[float] = None,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = False,
        checkpoint_subpath: str = "~/mava/",
    ):
        """Initializes the learner.
        Args:
            policy_networks: ...
            critic_networks: ...
            policy_optimizer: policy optimizer,
                This can be a single optimizer or an optimizer per agent key.
            critic_optimizer: critic optimizer,
            shared_weights: ...
            discount: discount to use for TD updates.
            dataset: dataset to learn from, whether fixed or from a replay buffer
                (see `acme.datasets.reverb.make_dataset` documentation).
            lambda_gae: ...
            clipping_espilon: ...
            entropy_cost: ...
            baseline_cost: ...
            max_abs_reward: ...
            max_gradient_norm: ...
            clipping: whether to clip gradients by global norm.
            counter: counter object used to keep track of steps.
            logger: logger object to be used by learner.
            checkpoint: boolean indicating whether to checkpoint the learner.
        """
        # Store agents.
        self._agents = agents
        self._agent_types = agent_types
        self._checkpoint = checkpoint

        # Store shared_weights.
        self._shared_weights = shared_weights

        # Store networks.
        self._observation_networks = observation_networks
        self._policy_networks = policy_networks
        self._critic_networks = critic_networks

        # Dictionary with network keys for each agent.
        self.agent_net_keys = {agent: agent for agent in self._agents}
        if self._shared_weights:
            self.agent_net_keys = {agent: agent.split("_")[0] for agent in self._agents}

        self.unique_net_keys = self._agent_types if shared_weights else self._agents

        # Create optimizers for different agent types.
        if not isinstance(policy_optimizer, dict):
            self._policy_optimizers: Dict[str, snt.Optimizer] = {}
            for agent in self.unique_net_keys:
                self._policy_optimizers[agent] = copy.deepcopy(policy_optimizer)
        else:
            self._policy_optimizers = policy_optimizer

        self._critic_optimizers: Dict[str, snt.Optimizer] = {}
        for agent in self.unique_net_keys:
            self._critic_optimizers[agent] = copy.deepcopy(critic_optimizer)

        # Expose the variables.
        policy_networks_to_expose = {}
        self._system_network_variables: Dict[str, Dict[str, snt.Module]] = {
            "critic": {},
            "policy": {},
        }
        for agent_key in self.unique_net_keys:
            policy_network_to_expose = snt.Sequential(
                [
                    self._observation_networks[agent_key],
                    self._policy_networks[agent_key],
                ]
            )
            policy_networks_to_expose[agent_key] = policy_network_to_expose
            self._system_network_variables["critic"][agent_key] = critic_networks[
                agent_key
            ].variables
            self._system_network_variables["policy"][
                agent_key
            ] = policy_network_to_expose.variables

        # Other trainer parameters.
        self._discount = discount
        self._entropy_cost = entropy_cost
        self._baseline_cost = baseline_cost
        self._lambda_gae = lambda_gae
        self._clipping_epsilon = clipping_epsilon

        # Dataset iterator
        self._iterator = dataset

        # Set up gradient clipping.
        if max_gradient_norm is not None:
            self._max_gradient_norm = tf.convert_to_tensor(max_gradient_norm)
        else:  # A very large number. Infinity results in NaNs.
            self._max_gradient_norm = tf.convert_to_tensor(1e10)

        # General learner book-keeping and loggers.
        self._counter = counter or counting.Counter()
        self._logger = logger or loggers.make_default_logger("trainer")

        # Create checkpointer
        self._system_checkpointer = {}
        if checkpoint:
            for agent_key in self.unique_net_keys:
                objects_to_save = {
                    "counter": self._counter,
                    "policy": self._policy_networks[agent_key],
                    "critic": self._critic_networks[agent_key],
                    "observation": self._observation_networks[agent_key],
                    "policy_optimizer": self._policy_optimizers,
                    "critic_optimizer": self._critic_optimizers,
                }

                subdir = os.path.join("trainer", agent_key)
                checkpointer = tf2_savers.Checkpointer(
                    time_delta_minutes=15,
                    directory=checkpoint_subpath,
                    objects_to_save=objects_to_save,
                    subdirectory=subdir,
                )
                self._system_checkpointer[agent_key] = checkpointer

        # Do not record timestamps until after the first learning step is done.
        # This is to avoid including the time it takes for actors to come online and
        # fill the replay buffer.
        self._timestamp: Optional[float] = None

    def _get_critic_feed(
        self,
        observations_trans: Dict[str, np.ndarray],
        agent: str,
    ) -> tf.Tensor:

        # Decentralised based
        observation_feed = observations_trans[agent]

        return observation_feed

    def _transform_observations(
        self, observation: Dict[str, np.ndarray]
    ) -> Dict[str, np.ndarray]:
        observation_trans = {}
        for agent in self._agents:
            agent_key = self.agent_net_keys[agent]
            observation_trans[agent] = self._observation_networks[agent_key](
                observation[agent].observation
            )
        return observation_trans

    @tf.function
    def _step(
        self,
    ) -> Dict[str, Dict[str, Any]]:

        # Get data from replay.
        inputs = next(self._iterator)

        self._forward(inputs)

        self._backward()

        # Log losses per agent
        return train_utils.map_losses_per_agent_ac(
            self.critic_losses, self.policy_losses
        )

    # Forward pass that calculates loss.
    def _forward(self, inputs: Any) -> None:
        # Convert to sequence data
        data = tf2_utils.batch_to_sequence(inputs.data)

        # Unpack input data as follows:
        observations, actions, rewards, discounts, extras = (
            data.observations,
            data.actions,
            data.rewards,
            data.discounts,
            data.extras,
        )

        # transform observation using observation networks
        observations_trans = self._transform_observations(observations)

        # Get log_probs.
        log_probs = extras["log_probs"]

        # Store losses.
        policy_losses: Dict[str, Any] = {}
        critic_losses: Dict[str, Any] = {}

        with tf.GradientTape(persistent=True) as tape:
            for agent in self._agents:

                action, reward, discount, behaviour_log_prob = (
                    actions[agent],
                    rewards[agent],
                    discounts[agent],
                    log_probs[agent],
                )

                actor_observation = observations_trans[agent]
                critic_observation = self._get_critic_feed(observations_trans, agent)

                # Chop off final timestep for bootstrapping value
                reward = reward[:-1]
                discount = discount[:-1]

                # Get agent network
                agent_key = agent.split("_")[0] if self._shared_weights else agent
                policy_network = self._policy_networks[agent_key]
                critic_network = self._critic_networks[agent_key]

                # Reshape inputs.
                dims = actor_observation.shape[:2]
                actor_observation = snt.merge_leading_dims(
                    actor_observation, num_dims=2
                )
                critic_observation = snt.merge_leading_dims(
                    critic_observation, num_dims=2
                )
                policy = policy_network(actor_observation)
                values = critic_network(critic_observation)

                # Reshape the outputs.
                policy = tfd.BatchReshape(policy, batch_shape=dims, name="policy")
                values = tf.reshape(values, dims, name="value")

                # Values along the sequence T.
                bootstrap_value = values[-1]
                state_values = values[:-1]

                # Generalized Return Estimation
                td_loss, td_lambda_extra = trfl.td_lambda(
                    state_values=state_values,
                    rewards=reward,
                    pcontinues=discount,
                    bootstrap_value=bootstrap_value,
                    lambda_=self._lambda_gae,
                    name="CriticLoss",
                )

                # Do not use the loss provided by td_lambda as they sum the losses over
                # the sequence length rather than averaging them.
                critic_loss = self._baseline_cost * tf.reduce_mean(
                    tf.square(td_lambda_extra.temporal_differences), name="CriticLoss"
                )

                # Compute importance sampling weights: current policy / behavior policy.
                log_rhos = policy.log_prob(action) - behaviour_log_prob
                importance_ratio = tf.exp(log_rhos)[:-1]
                clipped_importance_ratio = tf.clip_by_value(
                    importance_ratio,
                    1.0 - self._clipping_epsilon,
                    1.0 + self._clipping_epsilon,
                )

                # Generalized Advantage Estimation
                gae = tf.stop_gradient(td_lambda_extra.temporal_differences)
                mean, variance = tf.nn.moments(gae, axes=[0, 1], keepdims=True)
                normalized_gae = (gae - mean) / tf.sqrt(variance)

                policy_gradient_loss = tf.reduce_mean(
                    -tf.minimum(
                        tf.multiply(importance_ratio, normalized_gae),
                        tf.multiply(clipped_importance_ratio, normalized_gae),
                    ),
                    name="PolicyGradientLoss",
                )

                # Entropy regularization. Only implemented for categorical dist.
                try:
                    policy_entropy = tf.reduce_mean(policy.entropy())
                except NotImplementedError:
                    policy_entropy = tf.convert_to_tensor(0.0)

                entropy_loss = -self._entropy_cost * policy_entropy

                # Combine weighted sum of actor & entropy regularization.
                policy_loss = policy_gradient_loss + entropy_loss

                policy_losses[agent] = policy_loss
                critic_losses[agent] = critic_loss

        self.policy_losses = policy_losses
        self.critic_losses = critic_losses
        self.tape = tape

    # Backward pass that calculates gradients and updates network.
    def _backward(self) -> None:
        # Calculate the gradients and update the networks
        policy_losses = self.policy_losses
        critic_losses = self.critic_losses
        tape = self.tape

        for agent in self._agents:
            # Get agent_key.
            agent_key = agent.split("_")[0] if self._shared_weights else agent

            # Get trainable variables.
            policy_variables = (
                self._observation_networks[agent_key].trainable_variables
                + self._policy_networks[agent_key].trainable_variables
            )
            critic_variables = self._critic_networks[agent_key].trainable_variables

            # Get gradients.
            policy_gradients = tape.gradient(policy_losses[agent], policy_variables)
            critic_gradients = tape.gradient(critic_losses[agent], critic_variables)

            # Optionally apply clipping.
            critic_grads, critic_norm = tf.clip_by_global_norm(
                critic_gradients, self._max_gradient_norm
            )
            policy_grads, policy_norm = tf.clip_by_global_norm(
                policy_gradients, self._max_gradient_norm
            )

            # Apply gradients.
            self._critic_optimizers[agent_key].apply(critic_grads, critic_variables)
            self._policy_optimizers[agent_key].apply(policy_grads, policy_variables)

        train_utils.safe_del(self, "tape")

    def step(self) -> None:

        # Run the learning step.
        fetches = self._step()

        # Compute elapsed time.
        timestamp = time.time()
        elapsed_time = timestamp - self._timestamp if self._timestamp else 0
        self._timestamp = timestamp

        # Update our counts and record it.
        counts = self._counter.increment(steps=1, walltime=elapsed_time)
        fetches.update(counts)

        # Checkpoint and attempt to write the logs.
        if self._checkpoint:
            train_utils.checkpoint_networks(self._system_checkpointer)

        if self._logger:
            self._logger.write(fetches)

    def get_variables(self, names: Sequence[str]) -> Dict[str, Dict[str, np.ndarray]]:
        variables: Dict[str, Dict[str, np.ndarray]] = {}
        for network_type in names:
            variables[network_type] = {
                agent: tf2_utils.to_numpy(
                    self._system_network_variables[network_type][agent]
                )
                for agent in self.unique_net_keys
            }
        return variables


class CentralisedMAPPOTrainer(MAPPOTrainer):
    def __init__(
        self,
        agents: List[Any],
        agent_types: List[str],
        observation_networks: Dict[str, snt.Module],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        dataset: tf.data.Dataset,
        policy_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        critic_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        shared_weights: bool,
        discount: float = 0.99,
        lambda_gae: float = 1.0,
        entropy_cost: float = 0.0,
        baseline_cost: float = 1.0,
        clipping_epsilon: float = 0.2,
        max_gradient_norm: Optional[float] = None,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = False,
        checkpoint_subpath: str = "Checkpoints",
    ):

        super().__init__(
            agents=agents,
            agent_types=agent_types,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            observation_networks=observation_networks,
            dataset=dataset,
            shared_weights=shared_weights,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            discount=discount,
            lambda_gae=lambda_gae,
            entropy_cost=entropy_cost,
            baseline_cost=baseline_cost,
            clipping_epsilon=clipping_epsilon,
            max_gradient_norm=max_gradient_norm,
            counter=counter,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
        )

    def _get_critic_feed(
        self,
        observations_trans: Dict[str, np.ndarray],
        agent: str,
    ) -> tf.Tensor:
        # Centralised based
        observation_feed = tf.stack([x for x in observations_trans.values()], 2)

        return observation_feed



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import dm_env

from mava.utils.debugging.environments import TwoStepEnv, switch_game
from mava.utils.debugging.make_env import make_debugging_env
from mava.wrappers.debugging_envs import (
    DebuggingEnvWrapper,
    SwitchGameWrapper,
    TwoStepWrapper,
)


def make_environment(
    evaluation: bool,
    env_name: str = "simple_spread",
    action_space: str = "discrete",
    num_agents: int = 3,
    render: bool = False,
    return_state_info: bool = False,
    random_seed: Optional[int] = None,
) -> dm_env.Environment:

    assert action_space == "continuous" or action_space == "discrete"

    del evaluation

    if env_name == "two_step":
        environment = TwoStepEnv()
        environment = TwoStepWrapper(environment)
    elif env_name == "switch":
        """Creates a SwitchGame environment."""
        env_module_fn = switch_game.MultiAgentSwitchGame(num_agents=num_agents)
        environment_fn = SwitchGameWrapper(env_module_fn)
        return environment_fn
    else:
        """Creates a MPE environment."""
        env_module = make_debugging_env(env_name, action_space, num_agents)
        environment = DebuggingEnvWrapper(
            env_module, return_state_info=return_state_info
        )

        if random_seed and hasattr(environment, "seed"):
            environment.seed(random_seed)

    return environment



# Adapted from https://github.com/openai/multiagent-particle-envs.
# TODO (dries): Try using this class directly from PettingZoo and delete this file.

"""
Code for creating a multiagent environment with one of the scenarios listed
in ./scenarios/.
Can be called by using, for example:
    env = make_env('simple_speaker_listener')
After producing the env object, can be used similarly to an OpenAI gym
environment.
A policy using this environment must output actions in the form of a list
for all agents. Each element of the list should be a numpy array,
of size (env.world.dim_p + env.world.dim_c, 1). Physical actions precede
communication actions in this array. See environment.py for more details.
"""
from gym import Space

from . import scenarios as scenarios
from .environment import MultiAgentEnv


def make_debugging_env(
    scenario_name: str, action_space: Space, num_agents: int
) -> MultiAgentEnv:
    """
    Creates a MultiAgentEnv object as env. This can be used similar to a gym
    environment by calling env.reset() and env.step().
    Use env.render() to view the environment on the screen.
    Input:
        scenario_name   :   name of the scenario from ./scenarios/ to be Returns
                            (without the .py extension)
        benchmark       :   whether you want to produce benchmarking data
                            (usually only done during evaluation)
    Some useful env properties (see environment.py):
        .observation_space  :   Returns the observation space for each agent
        .action_space       :   Returns the action space for each agent
        .n                  :   Returns the number of Agents
    """

    # load scenario from script
    scenario = scenarios.load(scenario_name).Scenario()
    # create world
    world = scenario.make_world(num_agents)
    # create multiagent environment

    env = MultiAgentEnv(
        world,
        action_space,
        reset_callback=scenario.reset_world,
        reward_callback=scenario.reward,
        observation_callback=scenario.observation,
        done_callback=scenario.done,
    )
    return env



import collections
from typing import Any, Dict, List, Tuple, Union

import dm_env
import numpy as np
from dm_env import specs
from pettingzoo.utils.conversions import ParallelEnv
from pettingzoo.utils.env import AECEnv

# Need to install typing_extensions since we support pre python 3.8
from typing_extensions import TypedDict

from mava import types

SeqTimestepDict = TypedDict(
    "SeqTimestepDict",
    {"timestep": dm_env.TimeStep, "action": types.Action},
)


def generate_zeros_from_spec(spec: specs.Array) -> np.ndarray:
    return np.zeros(spec.shape, spec.dtype)


def convert_np_type(dtype: np.dtype, value: Union[int, float]) -> Union[int, float]:
    return np.dtype(dtype).type(value)


def parameterized_restart(
    reward: types.Reward,
    discount: types.Discount,
    observation: types.Observation,
) -> dm_env.TimeStep:
    """Returns a `TimeStep` with `step_type` set to `StepType.FIRST`.
    Differs from dm_env.restart, since reward and discount can be set to
    initial types."""
    return dm_env.TimeStep(dm_env.StepType.FIRST, reward, discount, observation)


"""Project single timestep to all agents."""


def broadcast_timestep_to_all_agents(
    timestep: dm_env.TimeStep, possible_agents: list
) -> dm_env.TimeStep:
    parallel_timestep = dm_env.TimeStep(
        observation={agent: timestep.observation for agent in possible_agents},
        reward={agent: timestep.reward for agent in possible_agents},
        discount={agent: timestep.discount for agent in possible_agents},
        step_type=timestep.step_type,
    )

    return parallel_timestep


"""Convert dict of seq timestep and actions to parallel"""


def convert_seq_timestep_and_actions_to_parallel(
    timesteps: Dict[str, SeqTimestepDict], possible_agents: list
) -> Tuple[dict, dm_env.TimeStep]:

    step_types = [timesteps[agent]["timestep"].step_type for agent in possible_agents]
    assert all(
        x == step_types[0] for x in step_types
    ), f"Step types should be identical - {step_types} "
    parallel_timestep = dm_env.TimeStep(
        observation={
            agent: timesteps[agent]["timestep"].observation for agent in possible_agents
        },
        reward={
            agent: timesteps[agent]["timestep"].reward for agent in possible_agents
        },
        discount={
            agent: timesteps[agent]["timestep"].discount for agent in possible_agents
        },
        step_type=step_types[0],
    )

    parallel_actions = {agent: timesteps[agent]["action"] for agent in possible_agents}

    return parallel_actions, parallel_timestep


def apply_env_wrapper_preprocessers(
    environment: Any,
    env_preprocess_wrappers: List,
) -> Any:
    # Currently only supports PZ envs.
    if isinstance(environment, ParallelEnv) or isinstance(environment, AECEnv):
        if env_preprocess_wrappers and isinstance(env_preprocess_wrappers, List):
            for (env_wrapper, params) in env_preprocess_wrappers:
                if params:
                    environment = env_wrapper(environment, **params)
                else:
                    environment = env_wrapper(environment)
    return environment


class RunningStatistics:
    """Helper class to comute running statistics such as
    the max, min, mean, variance and standard deviation of
    a specific quantity.
    """

    def __init__(self, label: str, queue_size: int = 100) -> None:

        self.queue: collections.deque = collections.deque(maxlen=queue_size)
        self._max = -float("inf")
        self._min = float("inf")

        self._mean = 0.0

        self._var = 0.0

        self._label = label

        self._raw = 0.0

    def push(self, x: float) -> None:
        self._raw = x
        self.queue.append(x)

        if x > self._max:
            self._max = x

        if x < self._min:
            self._min = x

        if len(self.queue) == 1:
            self._mean = x
            self._var = 0
        else:
            self._mean = np.mean(self.queue)
            self._var = np.var(self.queue)

    def max(self) -> float:
        return self._max

    def min(self) -> float:
        return self._min

    def mean(self) -> float:
        return self._mean

    def var(self) -> float:
        return self._var

    def std(self) -> float:
        return np.sqrt(self._var)

    def raw(self) -> float:
        return self._raw


# Adapted From https://github.com/DLR-RM/stable-baselines3/blob/237223f834fe9b8143ea24235d087c4e32addd2f/stable_baselines3/common/running_mean_std.py # noqa: E501
class RunningMeanStd(object):
    def __init__(self, epsilon: float = 1e-4, shape: Tuple[int, ...] = ()):
        """
        Calulates the running mean and std of a data stream
        https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
        :param epsilon: helps with arithmetic issues
        :param shape: the shape of the data stream's output
        """
        self.mean = np.zeros(shape, np.float64)
        self.var = np.ones(shape, np.float64)
        self.count = epsilon

    def update_batch(self, arr: np.ndarray) -> None:
        batch_mean = np.mean(arr, axis=0)
        batch_var = np.var(arr, axis=0)
        batch_count = arr.shape[0]
        self.update_from_moments(batch_mean, batch_var, batch_count)

    def update(self, arr: np.ndarray) -> None:
        batch_mean = np.mean(arr)
        batch_var = np.var(arr)
        batch_count = 1
        self.update_from_moments(batch_mean, batch_var, batch_count)

    def update_from_moments(
        self, batch_mean: np.ndarray, batch_var: np.ndarray, batch_count: int
    ) -> None:
        delta = batch_mean - self.mean
        tot_count = self.count + batch_count

        new_mean = self.mean + delta * batch_count / tot_count
        m_a = self.var * self.count
        m_b = batch_var * batch_count
        m_2 = (
            m_a
            + m_b
            + np.square(delta) * self.count * batch_count / (self.count + batch_count)
        )
        new_var = m_2 / (self.count + batch_count)

        new_count = batch_count + self.count

        self.mean = new_mean
        self.var = new_var
        self.count = new_count



# Adapted from https://github.com/openai/multiagent-particle-envs.
# TODO (dries): Try using this class directly from PettingZoo and delete this file.

from mava.utils.debugging.core import World


# defines scenario upon which the world is built
class BaseScenario(object):
    # create elements of the world
    def make_world(self, num_agents: int) -> World:
        raise NotImplementedError()

    # create initial conditions of the world
    def reset_world(self, world: World) -> None:
        raise NotImplementedError()



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List


def fully_connected_network_spec(
    agents_by_type: Dict[str, List[str]]
) -> Dict[str, List[str]]:
    """Creates network spec for fully connected agents by agent type"""
    network_spec: Dict[str, List[str]] = {}
    for agent_type, agents in agents_by_type.items():
        all_agents = agents
        for agent in agents:
            network_spec[agent] = all_agents
    return network_spec



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Executor implementations for VDN."""

from typing import Dict, Optional

import sonnet as snt
from acme.tf import variable_utils as tf2_variable_utils

from mava import adders
from mava.components.tf.modules.communication import BaseCommunicationModule
from mava.systems.tf.madqn.execution import MADQNFeedForwardExecutor
from mava.systems.tf.madqn.training import MADQNTrainer


class VDNFeedForwardExecutor(MADQNFeedForwardExecutor):
    """A feed-forward executor.
    An executor based on a feed-forward policy for each agent in the system
    which takes non-batched observations and outputs non-batched actions.
    It also allows adding experiences to replay and updating the weights
    from the policy on the learner.
    """

    def __init__(
        self,
        q_networks: Dict[str, snt.Module],
        action_selectors: Dict[str, snt.Module],
        trainer: MADQNTrainer,
        shared_weights: bool = True,
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
        communication_module: Optional[BaseCommunicationModule] = None,
        fingerprint: bool = False,
        evaluator: bool = False,
    ):
        """Initializes the executor.
        Args:
          policy_network: the policy to run for each agent in the system.
          shared_weights: specify if weights are shared between agent networks.
          adder: the adder object to which allows to add experiences to a
            dataset/replay buffer.
          variable_client: object which allows to copy weights from the trainer copy
            of the policies to the executor copy (in case they are separate).
        """
        super(VDNFeedForwardExecutor, self).__init__(
            q_networks=q_networks,
            action_selectors=action_selectors,
            shared_weights=shared_weights,
            adder=adder,
            variable_client=variable_client,
            communication_module=communication_module,
            fingerprint=fingerprint,
            trainer=trainer,
            evaluator=evaluator,
        )



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility function for building and launching launchpad programs."""

import functools
import inspect
import time
from typing import Any, Callable, Optional

import launchpad as lp
from absl import flags, logging
from acme.utils import counting

FLAGS = flags.FLAGS


def partial_kwargs(function: Callable[..., Any], **kwargs: Any) -> Callable[..., Any]:
    """Return a partial function application by overriding default keywords.
    This function is equivalent to `functools.partial(function, **kwargs)` but
    will raise a `ValueError` when called if either the given keyword arguments
    are not defined by `function` or if they do not have defaults.
    This is useful as a way to define a factory function with default parameters
    and then to override them in a safe way.
    Args:
      function: the base function before partial application.
      **kwargs: keyword argument overrides.
    Returns:
      A function.
    """
    # Try to get the argspec of our function which we'll use to get which keywords
    # have defaults.
    argspec = inspect.getfullargspec(function)

    # Figure out which keywords have defaults.
    if argspec.defaults is None:
        defaults = []
    else:
        defaults = argspec.args[-len(argspec.defaults) :]

    # Find any keys not given as defaults by the function.
    unknown_kwargs = set(kwargs.keys()).difference(defaults)

    # Raise an error
    if unknown_kwargs:
        error_string = "Cannot override unknown or non-default kwargs: {}"
        raise ValueError(error_string.format(", ".join(unknown_kwargs)))

    return functools.partial(function, **kwargs)


class StepsLimiter:
    """Process that terminates an experiment when `max_steps` is reached."""

    def __init__(
        self,
        counter: counting.Counter,
        max_steps: Optional[int],
        steps_key: str = "executor_steps",
    ):
        self._counter = counter
        self._max_steps = max_steps
        self._steps_key = steps_key

    def run(self) -> None:
        """Run steps limiter to terminate an experiment when max_steps is reached."""

        logging.info(
            "StepsLimiter: Starting with max_steps = %d (%s)",
            self._max_steps,
            self._steps_key,
        )
        while True:
            # Update the counts.
            counts = self._counter.get_counts()
            num_steps = counts.get(self._steps_key, 0)

            logging.info("StepsLimiter: Reached %d recorded steps", num_steps)

            if num_steps > self._max_steps:
                logging.info(
                    "StepsLimiter: Max steps of %d was reached, terminating",
                    self._max_steps,
                )
                lp.stop()

            # Don't spam the counter.
            time.sleep(10.0)



"""Generic executor implementations, using TensorFlow and Sonnet."""
from typing import Any, Dict, Optional, Tuple, Union

import dm_env
import sonnet as snt
import tensorflow as tf
import tensorflow_probability as tfp
from acme import types

# Internal imports.
from acme.tf import utils as tf2_utils
from acme.tf import variable_utils as tf2_variable_utils

from mava import adders, core
from mava.components.tf.modules.communication import BaseCommunicationModule

tfd = tfp.distributions


class FeedForwardExecutor(core.Executor):
    """A feed-forward executor.
    An executor based on a feed-forward policy for each agent in the system
    which takes non-batched observations and outputs non-batched actions.
    It also allows adding experiences to replay and updating the weights
    from the policy on the learner.
    """

    def __init__(
        self,
        policy_networks: Dict[str, snt.Module],
        shared_weights: bool = True,
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
    ):
        """Initializes the executor.
        Args:
          policy_network: the policy to run for each agent in the system.
          shared_weights: specify if weights are shared between agent networks.
          adder: the adder object to which allows to add experiences to a
            dataset/replay buffer.
          variable_client: object which allows to copy weights from the trainer copy
            of the policies to the executor copy (in case they are separate).
        """

        # Store these for later use.
        self._shared_weights = shared_weights
        self._policy_networks = policy_networks
        self._adder = adder
        self._variable_client = variable_client

    @tf.function
    def _policy(
        self, agent: str, observation: types.NestedTensor
    ) -> types.NestedTensor:

        # Add a dummy batch dimension and as a side effect convert numpy to TF.
        batched_observation = tf2_utils.add_batch_dim(observation)

        # index network either on agent type or on agent id
        agent_key = agent.split("_")[0] if self._shared_weights else agent

        # Compute the policy, conditioned on the observation.
        policy = self._policy_networks[agent_key](batched_observation)

        # Sample from the policy if it is stochastic.
        action = policy.sample() if isinstance(policy, tfd.Distribution) else policy

        return action

    def select_action(
        self, agent: str, observation: types.NestedArray
    ) -> Union[types.NestedArray, Tuple[types.NestedArray, types.NestedArray]]:
        # Pass the observation through the policy network.
        action = self._policy(agent, observation.observation)

        # TODO Mask actions here using observation.legal_actions
        # What happens in discrete vs cont case

        # Return a numpy array with squeezed out batch dimension.
        return tf2_utils.to_numpy_squeeze(action)

    def observe_first(
        self,
        timestep: dm_env.TimeStep,
        extras: Dict[str, types.NestedArray] = {},
    ) -> None:

        if self._adder:
            self._adder.add_first(timestep, extras)

    def observe(
        self,
        actions: Dict[str, types.NestedArray],
        next_timestep: dm_env.TimeStep,
        next_extras: Dict[str, types.NestedArray] = {},
    ) -> None:
        if self._adder:
            self._adder.add(actions, next_timestep, next_extras)

    def select_actions(
        self, observations: Dict[str, types.NestedArray]
    ) -> Union[
        Dict[str, types.NestedArray],
        Tuple[Dict[str, types.NestedArray], Dict[str, types.NestedArray]],
    ]:
        actions = {}
        for agent, observation in observations.items():
            # Pass the observation through the policy network.
            action = self._policy(agent, observation.observation)
            # TODO Mask actions here using observation.legal_actions
            # What happens in discrete vs cont case
            actions[agent] = tf2_utils.to_numpy_squeeze(action)

        # Return a numpy array with squeezed out batch dimension.
        return actions

    def update(self, wait: bool = False) -> None:
        if self._variable_client:
            self._variable_client.update(wait)


class RecurrentExecutor(core.Executor):
    """A recurrent Executor.
    An executor based on a recurrent policy for each agent in the system which
    takes non-batched observations and outputs non-batched actions, and keeps
    track of the recurrent state inside. It also allows adding experiences to
    replay and updating the weights from the policy on the learner.
    """

    def __init__(
        self,
        policy_networks: Dict[str, snt.RNNCore],
        shared_weights: bool = True,
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
        store_recurrent_state: bool = True,
    ):
        """Initializes the executor.
        Args:
          policy_networks: the (recurrent) policy to run for each agent in the system.
          shared_weights: specify if weights are shared between agent networks.
          adder: the adder object to which allows to add experiences to a
            dataset/replay buffer.
          variable_client: object which allows to copy weights from the trainer copy
            of the policies to the executor copy (in case they are separate).
          store_recurrent_state: Whether to pass the recurrent state to the adder.
        """
        # Store these for later use.
        self._policy_networks = policy_networks
        self._shared_weights = shared_weights
        self._adder = adder
        self._variable_client = variable_client
        self._store_recurrent_state = store_recurrent_state
        self._states: Dict[str, Any] = {}

    @tf.function
    def _policy(
        self,
        agent: str,
        observation: types.NestedTensor,
        state: types.NestedTensor,
    ) -> Union[
        Tuple[types.NestedTensor, types.NestedTensor],
        Tuple[types.NestedTensor, types.NestedTensor, types.NestedTensor],
    ]:

        # Add a dummy batch dimension and as a side effect convert numpy to TF.
        batched_observation = tf2_utils.add_batch_dim(observation)

        # index network either on agent type or on agent id
        agent_key = agent.split("_")[0] if self._shared_weights else agent

        # Compute the policy, conditioned on the observation.
        policy, new_state = self._policy_networks[agent_key](batched_observation, state)

        # Sample from the policy if it is stochastic.
        action = policy.sample() if isinstance(policy, tfd.Distribution) else policy

        return action, new_state

    def _update_state(self, agent: str, new_state: types.NestedArray) -> None:
        # Bookkeeping of recurrent states for the observe method.
        self._states[agent] = new_state

    def select_action(
        self, agent: str, observation: types.NestedArray
    ) -> types.NestedArray:

        # TODO Mask actions here using observation.legal_actions
        # What happens in discrete vs cont case

        # Initialize the RNN state if necessary.
        if self._states[agent] is None:
            # index network either on agent type or on agent id
            agent_key = agent.split("_")[0] if self._shared_weights else agent
            self._states[agent] = self._policy_networks[agent_key].initial_state(1)

        # Step the recurrent policy forward given the current observation and state.
        policy_output, new_state = self._policy(
            agent, observation.observation, self._states[agent]
        )

        # Bookkeeping of recurrent states for the observe method.
        self._update_state(agent, new_state)

        # Return a numpy array with squeezed out batch dimension.
        return tf2_utils.to_numpy_squeeze(policy_output)

    def observe_first(
        self,
        timestep: dm_env.TimeStep,
        extras: Optional[Dict[str, types.NestedArray]] = {},
    ) -> None:
        # Re-initialize the RNN state.
        for agent, _ in timestep.observation.items():
            # index network either on agent type or on agent id
            agent_key = agent.split("_")[0] if self._shared_weights else agent
            self._states[agent] = self._policy_networks[agent_key].initial_state(1)

        if self._adder is not None:
            numpy_states = {
                agent: tf2_utils.to_numpy_squeeze(_state)
                for agent, _state in self._states.items()
            }

            if extras:
                extras.update({"core_states": numpy_states})
                self._adder.add_first(timestep, extras)
            else:
                self._adder.add_first(timestep, numpy_states)

    def observe(
        self,
        actions: Dict[str, types.NestedArray],
        next_timestep: dm_env.TimeStep,
        next_extras: Optional[Dict[str, types.NestedArray]] = {},
    ) -> None:
        if not self._adder:
            return

        if not self._store_recurrent_state:
            if next_extras:
                self._adder.add(actions, next_timestep, next_extras)
            else:
                self._adder.add(actions, next_timestep)
            return

        numpy_states = {
            agent: tf2_utils.to_numpy_squeeze(_state)
            for agent, _state in self._states.items()
        }
        if next_extras:
            next_extras.update({"core_states": numpy_states})
            self._adder.add(actions, next_timestep, next_extras)
        else:
            self._adder.add(actions, next_timestep, numpy_states)

    def select_actions(
        self, observations: Dict[str, types.NestedArray]
    ) -> Union[
        Dict[str, types.NestedArray],
        Tuple[Dict[str, types.NestedArray], Dict[str, types.NestedArray]],
    ]:
        actions = {}
        for agent, observation in observations.items():
            # Step the recurrent policy forward given the current observation and state.
            actions[agent] = self.select_action(agent=agent, observation=observation)
        return actions

    def update(self, wait: bool = False) -> None:
        if self._variable_client:
            self._variable_client.update(wait)


class RecurrentCommExecutor(RecurrentExecutor):
    """Recurrent executor where agents can communicate."""

    def __init__(
        self,
        policy_networks: Dict[str, snt.RNNCore],
        communication_module: BaseCommunicationModule,
        shared_weights: bool = True,
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
        store_recurrent_state: bool = True,
    ):
        """Initializes the executor.
        TODO: Update docstring
        Args:
          policy_networks: the (recurrent) policy to run for each agent in the system.
          shared_weights: specify if weights are shared between agent networks.
          adder: the adder object to which allows to add experiences to a
            dataset/replay buffer.
          variable_client: object which allows to copy weights from the trainer copy
            of the policies to the executor copy (in case they are separate).
          store_recurrent_state: Whether to pass the recurrent state to the adder.
        """
        # Store these for later use.

        self._adder = adder
        self._variable_client = variable_client
        self._policy_networks = policy_networks
        self._states: Dict[str, Any] = {}
        self._messages: Dict[str, Any] = {}
        self._store_recurrent_state = store_recurrent_state
        self._communication_module = communication_module
        self._shared_weights = shared_weights

    def _sample_action(
        self, action_policy: types.NestedTensor, agent: str
    ) -> types.NestedTensor:
        # Sample from the policy if it is stochastic.
        return (
            action_policy.sample()
            if isinstance(action_policy, tfd.Distribution)
            else action_policy
        )

    def _process_message(
        self, observation: types.NestedTensor, message_policy: types.NestedTensor
    ) -> types.NestedTensor:
        # Only one agent can message at each timestep
        return message_policy

    @tf.function
    def _policy(
        self,
        agent: str,
        observation: types.NestedTensor,
        state: types.NestedTensor,
        message: types.NestedTensor,
    ) -> Tuple[types.NestedTensor, types.NestedTensor, types.NestedTensor]:

        # Add a dummy batch dimension and as a side effect convert numpy to TF.
        batched_observation = tf2_utils.add_batch_dim(observation)

        # index network either on agent type or on agent id
        agent_key = agent.split("_")[0] if self._shared_weights else agent

        # Compute the policy, conditioned on the observation.
        (action_policy, message_policy), new_state = self._policy_networks[agent_key](
            batched_observation, state, message
        )

        action = self._sample_action(action_policy, agent)

        message = self._process_message(observation, message_policy)

        return action, message, new_state

    def observe_first(
        self,
        timestep: dm_env.TimeStep,
        extras: Optional[Dict[str, types.NestedArray]] = {},
    ) -> None:

        # Re-initialize the RNN state.
        for agent, _ in timestep.observation.items():
            # index network either on agent type or on agent id
            agent_key = agent.split("_")[0] if self._shared_weights else agent
            self._states[agent] = self._policy_networks[agent_key].initial_state(1)
            self._messages[agent] = self._policy_networks[agent_key].initial_message(1)

        if self._adder is not None:
            numpy_states = {
                agent: tf2_utils.to_numpy_squeeze(_state)
                for agent, _state in self._states.items()
            }
            numpy_messages = {
                agent: tf2_utils.to_numpy_squeeze(_message)
                for agent, _message in self._messages.items()
            }
            if extras is not None:
                extras.update(
                    {
                        "core_states": numpy_states,
                        "core_messages": numpy_messages,
                    }
                )
                self._adder.add_first(timestep, extras)
            else:
                raise NotImplementedError("Why is extras None?")

    def observe(
        self,
        actions: Dict[str, types.NestedArray],
        next_timestep: dm_env.TimeStep,
        next_extras: Optional[Dict[str, types.NestedArray]] = {},
    ) -> None:
        if not self._adder:
            return

        if not self._store_recurrent_state:
            if next_extras:
                self._adder.add(actions, next_timestep, next_extras)
            else:
                self._adder.add(actions, next_timestep)
            return

        numpy_states = {
            agent: tf2_utils.to_numpy_squeeze(_state)
            for agent, _state in self._states.items()
        }
        numpy_messages = {
            agent: tf2_utils.to_numpy_squeeze(_message)
            for agent, _message in self._messages.items()
        }

        if next_extras:
            next_extras.update(
                {
                    "core_states": numpy_states,
                    "core_messages": numpy_messages,
                }
            )
            self._adder.add(actions, next_timestep, next_extras)
        else:
            next_extras = {
                "core_states": numpy_states,
                "core_messages": numpy_messages,
            }
            self._adder.add(actions, next_timestep, next_extras)

    def select_action(
        self,
        agent: str,
        observation: types.NestedArray,
    ) -> Tuple[types.NestedArray, types.NestedArray]:
        """Get actions and messages of specific agent"""

        # Initialize the RNN state if necessary.
        if self._states[agent] is None:
            self._states[agent] = self._networks[agent].initial_state(1)

        message_inputs = self._communication_module.process_messages(self._messages)
        # Step the recurrent policy forward given the current observation and state.
        policy_output, message, new_state = self._policy(
            agent,
            observation.observation,
            self._states[agent],
            message_inputs[agent],
        )

        # Bookkeeping of recurrent states for the observe method.
        self._states[agent] = new_state
        self._messages[agent] = message

        # Return a numpy array with squeezed out batch dimension.
        return tf2_utils.to_numpy_squeeze(policy_output)

    def select_actions(
        self, observations: Dict[str, types.NestedArray]
    ) -> Dict[str, types.NestedArray]:
        actions = {}
        for agent, observation in observations.items():
            actions[agent] = self._select_action(agent, observations)
        # Return a numpy array with squeezed out batch dimension.
        return actions



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional, Tuple

import dm_env
import numpy as np
import sonnet as snt
import tensorflow as tf
import tensorflow_probability as tfp
from acme import types

# Internal imports.
from acme.tf import utils as tf2_utils
from acme.tf import variable_utils as tf2_variable_utils

from mava import adders, core

tfd = tfp.distributions


class MAPPOFeedForwardExecutor(core.Executor):
    """A recurrent Executor for MAPPO.
    An executor based on a recurrent policy for each agent in the system which
    takes non-batched observations and outputs non-batched actions, and keeps
    track of the recurrent state inside. It also allows adding experiences to
    replay and updating the weights from the policy on the learner.
    """

    def __init__(
        self,
        policy_networks: Dict[str, snt.Module],
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
        shared_weights: bool = True,
    ):

        """Initializes the executor.
        Args:
          networks: the (recurrent) policy to run for each agent in the system.
          shared_weights: specify if weights are shared between agent networks.
          adder: the adder object to which allows to add experiences to a
            dataset/replay buffer.
          variable_client: object which allows to copy weights from the trainer copy
            of the policies to the executor copy (in case they are separate).
        """

        # Store these for later use.
        self._adder = adder
        self._variable_client = variable_client
        self._policy_networks = policy_networks
        self._shared_weights = shared_weights
        self._prev_log_probs: Dict[str, Any] = {}

    @tf.function
    def _policy(
        self,
        agent: str,
        observation: types.NestedTensor,
    ) -> Tuple[types.NestedTensor, types.NestedTensor]:
        # Index network either on agent type or on agent id.
        network_key = agent.split("_")[0] if self._shared_weights else agent

        # Add a dummy batch dimension and as a side effect convert numpy to TF.
        observation = tf2_utils.add_batch_dim(observation.observation)

        # Compute the policy, conditioned on the observation.
        policy = self._policy_networks[network_key](observation)

        # Sample from the policy and compute the log likelihood.
        action = policy.sample()
        log_prob = policy.log_prob(action)

        # Cast for compatibility with reverb.
        # sample() returns a 'int32', which is a problem.
        if isinstance(policy, tfp.distributions.Categorical):
            action = tf.cast(action, "int64")

        return log_prob, action

    def select_action(
        self, agent: str, observation: types.NestedArray
    ) -> types.NestedArray:

        # Step the recurrent policy/value network forward
        # given the current observation and state.
        self._prev_log_probs[agent], action = self._policy(agent, observation)

        # Return a numpy array with squeezed out batch dimension.
        action = tf2_utils.to_numpy_squeeze(action)

        # TODO(Kale-ab) : Remove. This is for debugging.
        if np.isnan(action).any():
            print(
                f"Value error- Log Probs:{self._prev_log_probs[agent]} Action: {action} "  # noqa: E501
            )

        return action

    def select_actions(
        self, observations: Dict[str, types.NestedArray]
    ) -> Dict[str, types.NestedArray]:

        actions = {}
        for agent, observation in observations.items():
            action = self.select_action(agent, observation)
            actions[agent] = action

        return actions

    def observe_first(
        self,
        timestep: dm_env.TimeStep,
        extras: Dict[str, types.NestedArray] = {},
    ) -> None:

        if self._adder:
            self._adder.add_first(timestep)

    def observe(
        self,
        actions: Dict[str, types.NestedArray],
        next_timestep: dm_env.TimeStep,
        next_extras: Dict[str, types.NestedArray] = {},
    ) -> None:

        if not self._adder:
            return

        next_extras.update({"log_probs": self._prev_log_probs})

        next_extras = tf2_utils.to_numpy_squeeze(next_extras)

        self._adder.add(actions, next_timestep, next_extras)

    def update(self, wait: bool = False) -> None:
        if self._variable_client:
            self._variable_client.update(wait)



import multiprocessing as mp
import os
import time
from typing import Any, Dict, List, Tuple

import numpy as np

from mava.utils.environments.RoboCup_env.robocup_utils.agent import (  # type: ignore # noqa: E501
    Agent as basic_agent,
)
from mava.utils.environments.RoboCup_env.robocup_utils.trainer import (  # type: ignore # noqa: E501
    Trainer,
)
from mava.utils.environments.RoboCup_env.robocup_utils.util_functions import (  # type: ignore # noqa: E501
    wait_for_next_observations,
)


# spawn an agent of team_name, with position
def spawn_agent(
    team_name: str, team_id: int, agent_id: int, num_players: int, port: int
) -> basic_agent:
    """
    Used to run an agent in a seperate physical process.
    """
    # return type of agent by position, construct
    a = basic_agent(
        teamname=team_name, team_id=team_id, agent_id=agent_id, num_players=num_players
    )
    a.connect("localhost", port)
    return a


def run_server_thread(game_setting: str, include_wait: bool, port: int) -> None:
    # ./usr/local/bin/rcssmonitor
    command = (
        "/usr/local/bin/rcssserver -server::coach_w_referee=on"
        " -server::half_time=99999999 -server::game_logging=false"
        " -server::game_log_dated=false"
        " -server::keepaway_log_dated=false"
        " -server::keepaway_logging=false"
        " -server::text_log_dated=false"
        " -server::port="
        + str(port)
        + " -server::coach_port="
        + str(port + 1)
        + " -server::olcoach_port="
        + str(port + 2)
    )

    if game_setting == "domain_randomisation" or game_setting == "fixed_opponent":
        command += " -server::drop_ball_time=60000"
        pass
    elif game_setting == "reward_shaping":
        pass
    else:
        raise NotImplementedError(
            "Should drop_ball_time be included for this game_setting?"
        )

    if include_wait:
        command += " -server::synch_mode=false"
    else:
        command += " -server::synch_mode=true"
    os.system(command)


def start_trainer(port: int) -> Trainer:
    trainer = Trainer()
    trainer.connect("localhost", port + 1)
    return trainer


def run_monitor_thread(port: int) -> None:
    # ./usr/local/bin/rcssmonitor
    os.system("/usr/local/bin/rcssmonitor --server-port=" + str(port))


def start_server(game_setting: str, include_wait: bool, port: int) -> None:
    # Wait for server to startup completely
    # print("mode: ", mode)
    at = mp.Process(target=run_server_thread, args=(game_setting, include_wait, port))
    at.daemon = True
    at.start()
    time.sleep(0.2)


def start_monitor(port: int) -> None:
    # Wait for monitor process to complete
    at = mp.Process(target=run_monitor_thread, args=(port,))
    at.daemon = True
    at.start()
    time.sleep(0.2)


def connect_agents(
    team_name: str,
    team_id: int,
    num_per_team: int,
    num_players: int,
    start_id: int,
    port: int,
) -> List:
    # spawn all agents as seperate processes for maximum processing efficiency
    agents = []
    for agent_id in range(start_id, start_id + num_per_team):
        agents.append(
            spawn_agent(
                team_name=team_name,
                team_id=team_id,
                agent_id=agent_id,
                num_players=num_players,
                port=port,
            )
        )
        time.sleep(0.01)

    print("Spawned %d agents." % len(agents))
    return agents


class RoboCup2D:
    def __init__(
        self,
        game_setting: str = "reward_shaping",
        include_wait: bool = False,
        team_names: List = ["Team1", "Team2"],
        players_per_team: List = [11, 11],
        render_game: bool = False,
        game_length: int = 6000,
        beta: float = 0.1,
        port: int = 6000,
    ):

        self.game_setting = game_setting
        if type(players_per_team) is not list:
            players_per_team = [players_per_team, players_per_team]

        self.players_per_team = players_per_team
        self.num_players = players_per_team[0] + players_per_team[1]

        assert (
            type(players_per_team) == list
            and 0 < players_per_team[0] <= 11
            and 0 <= players_per_team[1] <= 11
        )

        if game_setting == "domain_randomisation":
            self.game_diff = 0.0
            game_length = 100
        elif game_setting == "reward_shaping":
            game_length = 6000
        elif game_setting == "fixed_opponent":
            self.game_diff = 0.0
            game_length = 200
        else:
            raise NotImplementedError(
                "Game setting not implemented yet: ", game_setting
            )

        self.beta = 0.01

        # Start the server
        start_server(game_setting, include_wait, port)

        self.team_names = team_names

        # Connect trainer which is used to control the RoboCup server
        self.trainer = start_trainer(port)

        # Initial reset to move agent to correct position
        if game_setting == "single_agent_score":
            time.sleep(0.001)
            self.trainer.reset_game(
                players_per_team, team_names, game_setting, self.game_diff
            )

        self.game_length = game_length
        self.game_step = 0
        self.game_scores = [0, 0]

        self.end_step = 0

        self.tot_out_step = 0.0
        self.tot_in_step_no_wait = 0.0
        self.tot_server_wait = 0.0

        # Render the game on a monitor
        if render_game:
            start_monitor(port)

        self.id_dict = {"l": 0, "r": 1}

        # Connect teams
        self.agents = {}
        # self.agent_ids = []
        self.previous_scores = [0, 0]
        start_id = 0
        for t_i in range(2):
            if players_per_team[t_i] > 0:
                agents = connect_agents(
                    team_name=team_names[t_i],
                    team_id=t_i,
                    num_per_team=players_per_team[t_i],
                    num_players=self.num_players,
                    start_id=start_id,
                    port=port,
                )
                for agent in agents:
                    self.agents["player_" + str(agent.agent_id)] = agent
                start_id += players_per_team[t_i]

    def step(self, actions: Dict[str, Any]) -> Tuple[Any, Any, Any, Any]:
        # Do agent update
        # start_step = time.time()
        self.game_step += 1

        for agent_key, agent in self.agents.items():
            agent.do_action(actions[agent_key])

        # Send done to show that trainer is done.
        if self.trainer:
            self.trainer.send_done()

        # Wait for the environment to step and provide the next observations
        wait_for_next_observations([self.trainer])

        # Check if done with game
        done = self.game_step > self.game_length

        # Calculate rewards
        rewards = {}
        score_add = [
            self.trainer.wm.score_l - self.previous_scores[0],
            self.trainer.wm.score_r - self.previous_scores[1],
        ]
        self.previous_scores[0] = self.trainer.wm.score_l
        self.previous_scores[1] = self.trainer.wm.score_r
        self.game_scores[0] += score_add[0]
        self.game_scores[1] += score_add[1]

        # Reset the game if a goal is scored
        if (
            self.game_setting == "reward_shaping"
            and score_add[0] != 0
            or score_add[1] != 0
        ):
            self.trainer.reset_game(
                self.players_per_team,
                self.team_names,
                self.game_setting,
                reset_stamina=False,
            )

        wm = self.trainer.wm
        for agent_key, agent in self.agents.items():
            agent.do_action(actions[agent_key])
            team_id = agent.team_id
            opponent_id = (team_id + 1) % 2

            # Calculate rewards
            player = wm.players[agent.agent_id]

            if self.game_setting == "reward_shaping":
                if player["teamname"] == "Team_A":
                    destination_coords = wm.goals[0]["coords"]
                elif player["teamname"] == "Team_B":
                    destination_coords = wm.goals[1]["coords"]
                else:
                    raise ValueError("Unknown team: ", player["teamname"])

                ball_goal_dist = wm.euclidean_distance(
                    wm.ball["coords"], destination_coords
                )
                next_x = float(wm.ball["coords"][0]) + float(wm.ball["delta_coords"][0])
                next_y = float(wm.ball["coords"][1]) + float(wm.ball["delta_coords"][1])
                ball_goal_delta_dist = wm.euclidean_distance(
                    (next_x, next_y), destination_coords
                )
                ball_towards_goal = ball_goal_dist - ball_goal_delta_dist

                next_x = float(player["coords"][0]) + float(player["delta_coords"][0])
                next_y = float(player["coords"][1]) + float(player["delta_coords"][1])

                scored = score_add[team_id] - score_add[opponent_id]

                # TODO: Change back
                # if scored > 0.1:
                #     if self.beta < 0.001:
                #         self.beta = 0.0
                #         print("Beta minimum reached.")
                #     else:
                #         self.beta *= 0.99
                #         print("Beta: ", self.beta)

                self.beta = 0.01
                reward = (
                    scored + ball_towards_goal * self.beta
                )  # + player_towards_ball * self.beta + scored
            elif self.game_setting == "domain_randomisation":
                reward = score_add[team_id] - score_add[opponent_id]

                # # TODONE: Remove this. Masking zero rewards!
                # if reward < 0:
                #     reward = 0

            else:
                raise NotImplementedError("Unknown game setting: ", self.game_setting)

            rewards[agent_key] = np.array(reward, np.float32)

        return self.__get_latest_obs(), rewards, self.__get_state(), done

    def reset(self) -> Tuple[Any, Any, Any]:
        self.game_step = 0
        if self.game_setting == "domain_randomisation":
            goal_diff = self.game_scores[0] - self.game_scores[1]
            if goal_diff != 0:
                if self.game_diff < 1.0:
                    self.game_diff += 0.01
                    # print("Game difficulty increased: ", self.game_diff)
                else:
                    self.game_diff = 1.0
                    print("Max game difficulty surpassed!")
            elif self.game_diff < 1.0:
                if self.game_diff > 0.0:
                    self.game_diff -= 0.01

                if self.game_diff < 0.0:
                    # Sometimes the game_diff can be slightly smaller than zero
                    # which causes problems.
                    self.game_diff = 0.0
                    # print("Game difficulty decreased: ", self.game_diff)

            self.game_length = int(100 + self.game_diff * 900)
            # print("Reset game length: ", self.game_length)
            # TODO: Update alpha
            # self.alpha = 1 / (max_dist * game_length * (1 + beta))

            self.trainer.reset_game(
                self.players_per_team,
                self.team_names,
                self.game_setting,
                self.game_diff,
            )
        elif self.game_setting == "reward_shaping":
            self.trainer.reset_game(
                self.players_per_team, self.team_names, self.game_setting
            )
        else:
            raise NotImplementedError("Unknown game setting: ", self.game_setting)

        self.game_scores = [0, 0]

        # print("Resetting.")
        # Switch teams
        # TODO: Add team switching for better training. Also maybe shuffle agents in a
        # single team. Therefore change the agent each policy is controlling at every
        # reset step. Make sure that the dictionary stuff does not break.

        # Only shuffle if teams are equal for now.
        # if self.players_per_team[0] == self.players_per_team[1]:
        #     team1 = self.agents[0:self.players_per_team[0]]
        #     team2 = self.agents[self.players_per_team[0]:]
        #
        #     random.shuffle(team1)
        #     random.shuffle(team2)
        #
        #     self.agents = team2
        #     self.agents.extend(team1)
        rewards = [None] * len(self.agents)

        # Wait for the environment to step and provide the next observations
        wait_for_next_observations([self.trainer])
        return self.__get_latest_obs(), rewards, self.__get_state()

    def __get_state(self) -> Dict[str, Any]:
        # Return latest observations to agents
        state_dict = self.trainer.get_state_dict()
        # TODO: Return per agent states. The states should be mirrored. So a critic and
        # policy can train on only one side but execute on both sides.
        # state_dict
        state_dict["game_step"] = self.game_step
        state_dict["game_length"] = self.game_length
        return state_dict

    def __get_latest_obs(self) -> Dict[str, Any]:
        # Return latest observations to agents
        obs = {}
        for agent_key, agent in self.agents.items():
            obs_dict = agent.get_latest_observation()
            if obs_dict:
                obs_dict["game_step"] = self.game_step
                obs_dict["game_length"] = self.game_length

            obs[agent_key] = obs_dict

        return obs



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from typing import Any, Dict, Iterator, Optional, Type

import reverb
import sonnet as snt
from acme.utils import counting

from mava import core, types
from mava.components.tf.modules.communication import BaseCommunicationModule
from mava.components.tf.modules.exploration.exploration_scheduling import (
    LinearExplorationScheduler,
)
from mava.components.tf.modules.mixing import MonotonicMixing
from mava.components.tf.modules.stabilising import FingerPrintStabalisation
from mava.systems.tf.madqn.builder import MADQNBuilder, MADQNConfig
from mava.systems.tf.qmix import execution, training
from mava.wrappers import DetailedTrainerStatisticsWithEpsilon

# TODO Clean up documentation


@dataclasses.dataclass
class QMIXConfig(MADQNConfig):
    # TODO Match documentation with MADQNConfig once cleaned.
    """Configuration options for the QMIX system.
    Args:
            environment_spec: description of the actions, observations, etc.
            discount: discount to use for TD updates.
            batch_size: batch size for updates.
            prefetch_size: size to prefetch from replay.
            target_update_period: number of learner steps to perform before updating
              the target networks.
            min_replay_size: minimum replay size before updating.
            max_replay_size: maximum replay size.
            samples_per_insert: number of samples to take from replay for every insert
              that is made.
            n_step: number of steps to squash into a single transition.
            sigma: standard deviation of zero-mean, Gaussian exploration noise.
            clipping: whether to clip gradients by global norm.
            replay_table_name: string indicating what name to give the replay table."""


class QMIXBuilder(MADQNBuilder):
    """Builder for QMIX which constructs individual components of the system."""

    """Defines an interface for defining the components of an MARL system.
      Implementations of this interface contain a complete specification of a
      concrete MARL system. An instance of this class can be used to build an
      MARL system which interacts with the environment either locally or in a
      distributed setup.
      """

    def __init__(
        self,
        config: QMIXConfig,
        trainer_fn: Type[training.QMIXTrainer] = training.QMIXTrainer,
        executor_fn: Type[core.Executor] = execution.QMIXFeedForwardExecutor,
        mixer: Type[MonotonicMixing] = MonotonicMixing,
        extra_specs: Dict[str, Any] = {},
        exploration_scheduler_fn: Type[
            LinearExplorationScheduler
        ] = LinearExplorationScheduler,
        replay_stabilisation_fn: Optional[Type[FingerPrintStabalisation]] = None,
    ) -> None:
        """Args:
        _config: Configuration options for the QMIX system.
        _trainer_fn: Trainer module to use.
        """
        super(QMIXBuilder, self).__init__(
            config=config,
            trainer_fn=trainer_fn,
            executor_fn=executor_fn,
            extra_specs=extra_specs,
            exploration_scheduler_fn=exploration_scheduler_fn,
            replay_stabilisation_fn=replay_stabilisation_fn,
        )
        self._mixer = mixer

    def make_trainer(
        self,
        networks: Dict[str, Dict[str, snt.Module]],
        dataset: Iterator[reverb.ReplaySample],
        counter: Optional[counting.Counter] = None,
        logger: Optional[types.NestedLogger] = None,
        communication_module: Optional[BaseCommunicationModule] = None,
    ) -> core.Trainer:
        """Creates an instance of the trainer.
        Args:
          networks: struct describing the networks needed by the trainer; this can
            be specific to the trainer in question.
          dataset: iterator over samples from replay.
          counter: a Counter which allows for recording of counts (trainer steps,
            executor steps, etc.) distributed throughout the system.
          logger: Logger object for logging metadata.
        """
        agents = self._config.environment_spec.get_agent_ids()
        agent_types = self._config.environment_spec.get_agent_types()

        q_networks = networks["values"]
        target_q_networks = networks["target_values"]

        mixing_network = networks["mixing"]
        target_mixing_network = networks["target_mixing"]

        # Make epsilon scheduler
        exploration_scheduler = self._exploration_scheduler_fn(
            epsilon_min=self._config.epsilon_min,
            epsilon_decay=self._config.epsilon_decay,
        )

        # Check if we should use fingerprints
        fingerprint = True if self._replay_stabiliser_fn is not None else False

        # The learner updates the parameters (and initializes them).
        trainer = self._trainer_fn(  # type:ignore
            agents=agents,
            agent_types=agent_types,
            discount=self._config.discount,
            q_networks=q_networks,
            target_q_networks=target_q_networks,
            mixing_network=mixing_network,
            target_mixing_network=target_mixing_network,
            shared_weights=self._config.shared_weights,
            optimizer=self._config.optimizer,
            target_update_period=self._config.target_update_period,
            max_gradient_norm=self._config.max_gradient_norm,
            exploration_scheduler=exploration_scheduler,
            communication_module=communication_module,
            dataset=dataset,
            counter=counter,
            fingerprint=fingerprint,
            logger=logger,
            checkpoint=self._config.checkpoint,
            checkpoint_subpath=self._config.checkpoint_subpath,
        )

        trainer = DetailedTrainerStatisticsWithEpsilon(trainer)  # type:ignore

        return trainer
